# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import logging
import datetime
import pathlib
from typing import List
from typing import Dict
from typing import Optional

import tzlocal
import web
from web.utils import Storage

from ooresults.handler import model
from ooresults.plugins.imports.entries import text
from ooresults.plugins import oe2003
from ooresults.plugins import iof_entry_list
from ooresults.plugins import iof_result_list
from ooresults.repo import result_type
from ooresults.repo import start_type
from ooresults.repo.result_type import ResultStatus
from ooresults.repo.repo import EventNotFoundError
from ooresults.repo.repo import ConstraintError
from ooresults.utils.globals import t_globals


templates = pathlib.Path(__file__).resolve().parent.parent / "templates"
render = web.template.render(templates, globals=t_globals)


def update(event_id: int):
    entry_list = model.get_entries(event_id)
    event = list(model.get_event(event_id))
    if event == []:
        return render.entries_table({}, entry_list)
    else:
        return render.entries_table(event[0], entry_list)


class Update:
    def POST(self):
        """Update data"""
        data = web.input()
        event_id = int(data.event_id) if data.event_id != "" else -1
        return update(event_id)


class Import:
    def POST(self):
        """Import entries"""
        data = web.input()
        event_id = int(data.event_id) if data.event_id != "" else -1
        try:
            event = {}
            if data.entr_import == "entr.import.1":
                event, entries = iof_entry_list.parse_entry_list(data.browse1)
                model.import_entries(event_id=event_id, entries=entries)
            elif data.entr_import == "entr.import.2":
                event, entries = iof_result_list.parse_result_list(data.browse2)
                model.import_entries(event_id=event_id, entries=entries)
            elif data.entr_import == "entr.import.3":
                event = model.get_event(id=event_id)[0]
                entries = oe2003.parse(content=data.browse3)

                tz = tzlocal.get_localzone()
                for e in entries:
                    start_time = e["start"].start_time
                    if start_time is not None:
                        e["start"].start_time = datetime.datetime.combine(
                            event.date, start_time.time(), tzinfo=tz
                        )
                        print("StartTime(start):", e["start"].start_time)
                    start_time = e["result"].start_time
                    if start_time is not None:
                        e["result"].start_time = datetime.datetime.combine(
                            event.date, start_time.time(), tzinfo=tz
                        )
                        e["result"].punched_start_time = e["result"].start_time
                        print("StartTime(result):", e["result"].start_time)
                    finish_time = e["result"].finish_time
                    if finish_time is not None:
                        e["result"].finish_time = datetime.datetime.combine(
                            event.date, finish_time.time(), tzinfo=tz
                        )
                        e["result"].punched_finish_time = e["result"].finish_time
                        print("FinishTime(result):", e["result"].finish_time)
                    for i in e["result"].split_times:
                        if i.punch_time is not None:
                            i.punch_time = datetime.datetime.combine(
                                event.date, i.punch_time.time(), tzinfo=tz
                            )

                model.import_entries(event_id=event_id, entries=entries)
            elif data.entr_import == "entr.import.4":
                entries = text.parse(content=data.browse4)
                model.import_entries(event_id=event_id, entries=entries)

        except EventNotFoundError:
            raise web.conflict("No event selected or event deleted")

        except Exception as e:
            raise web.Conflict(str(e))

        return update(event_id)


class Export:
    def POST(self):
        """Export entries"""
        data = web.input()
        event_id = int(data.event_id) if data.event_id != "" else -1
        try:
            if data.entr_export == "entr.export.1":
                event = model.get_event(id=event_id)
                entry_list = model.get_entries(event_id=event_id)
                content = iof_entry_list.create_entry_list(event[0], entry_list)
            elif data.entr_export == "entr.export.2":
                event = model.get_event(id=event_id)
                entry_list = model.get_entries(event_id=event_id)
                content = iof_result_list.create_result_list(event[0], entry_list)
            elif data.entr_export == "entr.export.3":
                class_list = model.get_classes(event_id=event_id)
                entry_list = model.get_entries(event_id=event_id)
                content = oe2003.create(entry_list, list(class_list))

        except KeyError:
            raise web.conflict("Entry deleted")
        except:
            logging.exception("Internal server error")
            raise

        return content


class Add:
    def parse_start_time(
        self, item: str, event_date: datetime.date
    ) -> Optional[datetime.datetime]:
        if item != "":
            format = "%H:%M:%S" if item.count(":") == 2 else "%M:%S"
            tz = tzlocal.get_localzone()
            dt = datetime.datetime.combine(
                date=event_date,
                time=datetime.datetime.strptime(item, format).time(),
                tzinfo=tz,
            )
            print(">>> ", dt)
            return dt
        else:
            return None

    def POST(self):
        """Add or edit entry"""
        data = web.input()
        print(data)
        event_id = int(data.event_id) if data.event_id != "" else -1
        event = model.get_event(id=event_id)[0]

        try:
            entered_start_time = self.parse_start_time(data.start_time, event.date)

            fields = {}
            for i in range(len(event.fields)):
                name = "f" + str(i)
                if name in data:
                    fields[i] = data[name]

            model.add_or_update_entry(
                id=int(data.id) if data.id != "" else None,
                event_id=event_id,
                competitor_id=int(data.competitor_id)
                if data.competitor_id != ""
                else None,
                first_name=data.first_name,
                last_name=data.last_name,
                gender=data.gender,
                year=int(data.year) if data.year != "" else None,
                class_id=int(data.class_),
                club_id=int(data.club) if data.club != "" else None,
                not_competing="not_competing" in data and data.not_competing == "true",
                chip=data.chip,
                fields=fields,
                status=ResultStatus(int(data.status)),
                start_time=entered_start_time,
                result_id=int(data.get("result", ""))
                if data.get("result", "") != ""
                else None,
            )

        except EventNotFoundError:
            raise web.conflict("No event selected or event deleted")
        except ConstraintError as e:
            raise web.conflict(str(e))
        except KeyError:
            raise web.conflict("Entry deleted")
        except:
            logging.exception("Internal server error")
            raise

        return update(event_id)


class FillEditForm:
    def collect_not_assigned_si_results(self, entries) -> List[Dict]:
        results = []
        for e in entries:
            if e.last_name is None:
                last_punch = e.result.finish_time
                if last_punch is None:
                    for s in reversed(e.result.split_times):
                        if s.punch_time is not None:
                            last_punch = s.punch_time
                            break
                    else:
                        if e.result.start_time is not None:
                            last_punch = e.result.start_time

                if last_punch is None:
                    punch_time = "--:--:--"
                else:
                    punch_time = last_punch.strftime("%H:%M:%S")

                results.append({"key": e.id, "value": f"{punch_time}   --   {e.chip}"})
        return results

    def POST(self):
        """Query data to fill add or edit form"""
        data = web.input()
        event_id = int(data.event_id) if data.event_id != "" else -1
        event = model.get_event(id=event_id)[0]
        try:
            results = []
            if data.id == "":
                entry = Storage(
                    {
                        "id": "",
                        "competitor_id": "",
                        "first_name": "",
                        "last_name": "",
                        "class_id": "",
                        "class_name": "",
                        "club_id": "",
                        "club_name": "",
                        "gender": "",
                        "year": None,
                        "not_competing": False,
                        "chip": "",
                        "fields": {},
                        "result": result_type.PersonRaceResult(),
                        "start": start_type.PersonRaceStart(),
                    }
                )
            else:
                entry = model.get_entry(int(data.id))[0]
                if entry.result is not None and entry.result.has_punches():
                    results += [{"key": -1, "value": "Remove result"}]

            entries = model.get_entries(event_id=event_id)
            results += self.collect_not_assigned_si_results(entries)

            classes = model.get_classes(event_id=event_id)
            clubs = model.get_clubs()
        except EventNotFoundError:
            raise web.conflict("No event selected or event deleted")
        except KeyError:
            raise web.conflict("Entry deleted")
        except:
            logging.exception("Internal server error")
            raise

        return render.add_entry(entry, classes, clubs, results, event.fields)


class FillCompetitorsForm:
    def POST(self):
        """Query data to fill add or edit form"""
        data = web.input()
        competitors = model.get_competitors()

        return render.add_entry_competitors(competitors)


class FillResultForm:
    def POST(self):
        """Query data to fill result form"""
        data = web.input()
        event_id = int(data.event_id) if data.event_id != "" else -1
        try:
            entry = model.get_entry(int(data.id))[0]
        except EventNotFoundError:
            raise web.conflict("No event selected or event deleted")
        except KeyError:
            raise web.conflict("Entry deleted")
        except:
            logging.exception("Internal server error")
            raise

        return render.add_entry_result(entry)


class Delete:
    def POST(self):
        """Delete entry"""
        data = web.input()
        event_id = int(data.event_id) if data.event_id != "" else -1
        model.delete_entry(int(data.id))
        return update(event_id=event_id)



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import io
import sys
import copy
import logging
import pickle
import sqlite3
import datetime
from typing import Dict
from typing import List
from typing import Optional

import web
import web.db

from ooresults.repo.repo import Repo
from ooresults.repo.repo import ClassUsedError
from ooresults.repo.repo import CourseUsedError
from ooresults.repo.repo import ClubUsedError
from ooresults.repo.repo import CompetitorUsedError
from ooresults.repo.repo import EventNotFoundError
from ooresults.repo.repo import ConstraintError
from ooresults.repo.repo import OperationalError
from ooresults.repo.repo import TransactionMode
from ooresults.repo.update import update_tables
from ooresults.repo.class_params import ClassParams
from ooresults.repo import result_type
from ooresults.repo import series_type
from ooresults.repo import start_type


class UnpicklerZoneInfo(pickle.Unpickler):
    def find_class(self, module: str, name: str):
        if name == "ZoneInfo":
            if sys.version_info.major >= 3 and sys.version_info.minor >= 9:
                if module == "backports.zoneinfo":
                    module = "zoneinfo"
            else:
                if module == "zoneinfo":
                    module = "backports.zoneinfo"

        return pickle.Unpickler.find_class(self, module, name)


class SqliteRepo(Repo):
    def __init__(self, db: str = "ooresults.sqlite"):
        self.db = web.database(dbn="sqlite", db=db)

        # sqlite3.register_adapter(bool, int)
        # sqlite3.register_converter("BOOLEAN", lambda v: v != '0')

        values = list(
            self.db.query(
                "SELECT name FROM sqlite_schema WHERE type="
                + web.db.sqlquote("table")
                + " AND name="
                + web.db.sqlquote("version")
                + ";"
            )
        )

        if not values:
            # create and initialize database tables
            self.db.query("CREATE TABLE version (value INTEGER PRIMARY KEY);")
            self.db.query(
                "CREATE TABLE classes ("
                "id INTEGER PRIMARY KEY,"
                "event_id INTEGER NOT NULL,"
                "name TEXT NOT NULL,"
                "short_name TEXT,"
                "course_id INTEGER,"
                "params BLOB NOT NULL,"
                "UNIQUE (event_id, name)"
                ");"
            )
            self.db.query(
                "CREATE TABLE courses ("
                "id INTEGER PRIMARY KEY,"
                "event_id INTEGER NOT NULL,"
                "name TEXT NOT NULL,"
                "length FLOAT,"
                "climb FLOAT,"
                "controls BLOB NOT NULL,"
                "UNIQUE (event_id, name)"
                ");"
            )
            self.db.query(
                "CREATE TABLE clubs (" "id INTEGER PRIMARY KEY," "name TEXT UNIQUE" ");"
            )
            self.db.query(
                "CREATE TABLE competitors ("
                "id INTEGER PRIMARY KEY,"
                "first_name TEXT NOT NULL,"
                "last_name TEXT NOT NULL,"
                "club_id INTEGER,"
                "gender TEXT,"
                "year INTEGER,"
                "chip TEXT,"
                "UNIQUE (first_name, last_name)"
                ");"
            )
            self.db.query(
                "CREATE TABLE events ("
                "id INTEGER PRIMARY KEY,"
                "name TEXT NOT NULL UNIQUE,"
                "date TEXT,"
                "key TEXT UNIQUE,"
                "publish BOOL NOT NULL,"
                "series TEXT,"
                "fields BLOB NOT NULL"
                ");"
            )
            self.db.query(
                "CREATE TABLE entries ("
                "id INTEGER PRIMARY KEY,"
                "event_id INTEGER NOT NULL,"
                "competitor_id INTEGER,"
                "class_id INTEGER,"
                "club_id INTEGER,"
                "not_competing BOOL NOT NULL,"
                "result BLOB NOT NULL,"
                "start BLOB NOT NULL,"
                "chip TEXT,"
                "fields BLOB NOT NULL,"
                "UNIQUE (event_id, competitor_id)"
                ");"
            )
            self.db.query(
                "CREATE TABLE settings ("
                "name TEXT NOT NULL,"
                "nr_of_best_results INTEGER,"
                "mode TEXT NOT NULL,"
                "maximum_points INTEGER NOT NULL,"
                "decimal_places INTEGER NOT NULL"
                ");"
            )

            t = self.db.transaction()
            self.db.ctx.db.execute("BEGIN EXCLUSIVE TRANSACTION;")
            self.db.insert("version", value=update_tables.VERSION)
            t.commit()
            logging.info(f"DB version initialized to {update_tables.VERSION}")

        update_tables.update_tables(db=self.db, path=db)

    def start_transaction(self, mode: TransactionMode = TransactionMode.DEFERRED):
        self.t = self.db.transaction()
        try:
            self.db.ctx.db.execute(f"BEGIN {mode.value} TRANSACTION;")
        except sqlite3.OperationalError as e:
            raise OperationalError(str(e))

    def commit(self):
        self.t.commit()

    def rollback(self):
        self.t.rollback()

    def get_classes(self, event_id: int):
        values = self.db.query(
            "SELECT "
            "classes.id,"
            "classes.name,"
            "classes.short_name,"
            "courses.id,"
            "courses.name,"
            "classes.params "
            "FROM classes "
            "LEFT JOIN courses ON classes.course_id = courses.id "
            "WHERE classes.event_id = " + web.db.sqlquote(event_id) + " "
            "ORDER BY classes.name ASC;"
        )
        values.names[values.names.index("id", 1)] = "course_id"
        values.names[values.names.index("name", 2)] = "course"
        values = list(values)
        for c in values:
            c.params = UnpicklerZoneInfo(io.BytesIO(c.params)).load()
        return values

    def get_class(self, id: int):
        values = list(self.db.where("classes", id=id))
        for c in values:
            c.params = UnpicklerZoneInfo(io.BytesIO(c.params)).load()
        return values

    def add_class(
        self,
        event_id: int,
        name: str,
        short_name: Optional[str],
        course_id: Optional[int],
        params: ClassParams,
    ) -> int:
        if self.get_event(id=event_id) == []:
            raise EventNotFoundError
        try:
            return self.db.insert(
                "classes",
                event_id=event_id,
                name=name,
                short_name=short_name,
                course_id=course_id,
                params=pickle.dumps(params),
            )
        except sqlite3.IntegrityError:
            raise ConstraintError("Class already exist")

    def update_class(
        self,
        id: int,
        name: str,
        short_name: Optional[str],
        course_id: Optional[int],
        params: ClassParams,
    ):
        try:
            nr_of_rows = self.db.update(
                "classes",
                where="id=" + web.db.sqlquote(id),
                name=name,
                short_name=short_name,
                course_id=course_id,
                params=pickle.dumps(params),
            )
            if nr_of_rows == 0:
                raise KeyError
        except sqlite3.IntegrityError:
            raise ConstraintError("Class already exist")

    def delete_classes(self, event_id: int):
        if self.db.where("entries", event_id=event_id).first() is None:
            self.db.delete("classes", where="event_id=" + web.db.sqlquote(event_id))
        else:
            raise ClassUsedError

    def delete_class(self, id: int):
        if self.db.where("entries", class_id=id).first() is None:
            self.db.delete("classes", where="id=" + web.db.sqlquote(id))
        else:
            raise ClassUsedError

    def get_courses(self, event_id: int):
        data = list(
            self.db.select(
                "courses",
                where="event_id=" + web.db.sqlquote(event_id),
                order="name ASC",
            )
        )
        for c in data:
            c.controls = UnpicklerZoneInfo(io.BytesIO(c.controls)).load()
        return data

    def get_course(self, id: int):
        data = list(self.db.where("courses", id=id))
        for c in data:
            c.controls = UnpicklerZoneInfo(io.BytesIO(c.controls)).load()
        return data

    def add_course(
        self,
        event_id: int,
        name: str,
        length: Optional[float],
        climb: Optional[float],
        controls: List[str],
    ) -> int:
        if self.get_event(id=event_id) == []:
            raise EventNotFoundError
        try:
            return self.db.insert(
                "courses",
                event_id=event_id,
                name=name,
                length=length,
                climb=climb,
                controls=pickle.dumps(controls),
            )
        except sqlite3.IntegrityError:
            raise ConstraintError("Course already exist")

    def update_course(
        self,
        id: int,
        name: str,
        length: Optional[float],
        climb: Optional[float],
        controls: List[str],
    ):
        try:
            nr_of_rows = self.db.update(
                "courses",
                where="id=" + web.db.sqlquote(id),
                name=name,
                length=length,
                climb=climb,
                controls=pickle.dumps(controls),
            )
            if nr_of_rows == 0:
                raise KeyError
        except sqlite3.IntegrityError:
            raise ConstraintError("Course already exist")

    def delete_courses(self, event_id: int):
        if (
            self.db.select(
                "classes",
                where="event_id="
                + web.db.sqlquote(event_id)
                + " and course_id is not null",
            ).first()
            is None
        ):
            self.db.delete("courses", where="event_id=" + web.db.sqlquote(event_id))
        else:
            raise CourseUsedError

    def delete_course(self, id: int):
        if self.db.where("classes", course_id=id).first() is None:
            self.db.delete("courses", where="id=" + web.db.sqlquote(id))
        else:
            raise CourseUsedError

    def get_clubs(self):
        return self.db.select("clubs", order="name ASC")

    def get_club(self, id: int):
        return self.db.where("clubs", id=id)

    def add_club(self, name: str) -> int:
        try:
            return self.db.insert("clubs", name=name)
        except sqlite3.IntegrityError:
            raise ConstraintError("Club already exist")

    def update_club(self, id: int, name: str):
        try:
            nr_of_rows = self.db.update(
                "clubs", where="id=" + web.db.sqlquote(id), name=name
            )
            if nr_of_rows == 0:
                raise KeyError
        except sqlite3.IntegrityError:
            raise ConstraintError("Club already exist")

    def delete_club(self, id: int):
        if (
            self.db.where("competitors", club_id=id).first() is None
            and self.db.where("entries", club_id=id).first() is None
        ):
            self.db.delete("clubs", where="id=" + web.db.sqlquote(id))
        else:
            raise ClubUsedError

    def get_competitors(self):
        values = self.db.query(
            "SELECT "
            "competitors.id,"
            "competitors.first_name,"
            "competitors.last_name,"
            "competitors.gender,"
            "competitors.year,"
            "competitors.chip,"
            "clubs.id,"
            "clubs.name "
            "FROM competitors "
            "LEFT JOIN clubs ON competitors.club_id = clubs.id "
            "ORDER BY competitors.last_name ASC, competitors.first_name ASC;"
        )
        values.names[values.names.index("id", 1)] = "club_id"
        values.names[values.names.index("name")] = "club_name"
        return values

    def get_competitor(self, id):
        return self.db.where("competitors", id=id)

    def get_competitor_by_name(self, first_name: str, last_name: str):
        return self.db.where("competitors", first_name=first_name, last_name=last_name)

    def add_competitor(
        self,
        first_name: str,
        last_name: str,
        club_id: Optional[int],
        gender,
        year: Optional[int],
        chip,
    ) -> int:
        try:
            return self.db.insert(
                "competitors",
                first_name=first_name,
                last_name=last_name,
                club_id=club_id,
                gender=gender,
                year=year,
                chip=chip,
            )
        except sqlite3.IntegrityError:
            raise ConstraintError("Competitor already exist")

    def update_competitor(
        self,
        id: int,
        first_name: str,
        last_name: str,
        club_id: Optional[int],
        gender,
        year: Optional[int],
        chip,
    ) -> None:
        try:
            nr_of_rows = self.db.update(
                "competitors",
                where="id=" + web.db.sqlquote(id),
                first_name=first_name,
                last_name=last_name,
                club_id=club_id,
                gender=gender,
                year=year,
                chip=chip,
            )
            if nr_of_rows == 0:
                raise KeyError
        except sqlite3.IntegrityError:
            raise ConstraintError("Competitor already exist")

    def delete_competitor(self, id: int):
        if self.db.where("entries", competitor_id=id).first() is None:
            self.db.delete("competitors", where="id=" + web.db.sqlquote(id))
        else:
            raise CompetitorUsedError

    def import_competitors(self, competitors):
        list_of_competitors = []
        for c in competitors:
            for clb in self.get_clubs():
                if clb.name == c["club"]:
                    break
            else:
                self.add_club(c["club"])
            for clb in self.get_clubs():
                if clb.name == c["club"]:
                    break

            competitor = self.get_competitor_by_name(
                first_name=c["first_name"], last_name=c["last_name"]
            ).first()
            if competitor is None:
                list_of_competitors.append(
                    {
                        "last_name": c["last_name"],
                        "first_name": c["first_name"],
                        "club_id": clb.id,
                        "gender": c["gender"] if "gender" in c else "",
                        "year": c["year"] if "year" in c else "",
                        "chip": c["chip"] if "chip" in c else "",
                    }
                )
            else:
                self.update_competitor(
                    id=competitor.id,
                    first_name=competitor.first_name,
                    last_name=competitor.last_name,
                    club_id=clb.id
                    if competitor.club_id is None
                    else competitor.club_id,
                    gender=c["gender"]
                    if competitor.gender == "" and "gender" in c
                    else competitor.gender,
                    year=c["year"]
                    if competitor.year == "" and "year" in c
                    else competitor.year,
                    chip=c["chip"]
                    if competitor.chip == "" and "chip" in c
                    else competitor.chip,
                )

        self.db.supports_multiple_insert = True
        for i in range(0, len(competitors), 25):
            self.db.multiple_insert(
                "competitors", list_of_competitors[i : min(i + 25, len(competitors))]
            )

    def get_entries(self, event_id: int):
        values = self.db.query(
            "SELECT "
            "entries.id,"
            "entries.event_id,"
            "competitors.id,"
            "competitors.first_name,"
            "competitors.last_name,"
            "competitors.gender,"
            "competitors.year,"
            "entries.class_id,"
            "classes.name,"
            "entries.not_competing,"
            "entries.chip,"
            "entries.fields,"
            "entries.result,"
            "entries.start,"
            "entries.club_id,"
            "clubs.name "
            "FROM entries "
            "LEFT JOIN competitors ON entries.competitor_id = competitors.id "
            "LEFT JOIN classes ON entries.class_id = classes.id "
            "LEFT JOIN clubs ON entries.club_id = clubs.id "
            "WHERE entries.event_id = " + web.db.sqlquote(event_id) + " "
            "ORDER BY competitors.last_name ASC, competitors.first_name ASC, entries.chip ASC;"
        )
        values.names[values.names.index("id", 1)] = "competitor_id"
        values.names[values.names.index("name")] = "class_"
        values.names[values.names.index("name")] = "club"
        values = list(values)
        for e in values:
            e.fields = UnpicklerZoneInfo(io.BytesIO(e.fields)).load()
            e.result = UnpicklerZoneInfo(io.BytesIO(e.result)).load()
            e.start = UnpicklerZoneInfo(io.BytesIO(e.start)).load()
            e.not_competing = bool(e.not_competing)
        return values

    def get_entry(self, id: int):
        values = self.db.query(
            "SELECT "
            "entries.id,"
            "entries.event_id,"
            "competitors.id,"
            "competitors.first_name,"
            "competitors.last_name,"
            "competitors.gender,"
            "competitors.year,"
            "classes.id,"
            "classes.name,"
            "clubs.id,"
            "clubs.name,"
            "entries.not_competing,"
            "entries.chip,"
            "entries.fields,"
            "entries.result,"
            "entries.start "
            "FROM entries "
            "LEFT JOIN competitors ON entries.competitor_id = competitors.id "
            "LEFT JOIN classes ON entries.class_id = classes.id "
            "LEFT JOIN clubs ON entries.club_id = clubs.id "
            "WHERE entries.id = " + web.db.sqlquote(id) + ";"
        )
        values.names[values.names.index("id", 1)] = "competitor_id"
        values.names[values.names.index("id", 1)] = "class_id"
        values.names[values.names.index("id", 1)] = "club_id"
        values.names[values.names.index("name")] = "class_"
        values.names[values.names.index("name")] = "club"
        values = list(values)
        for e in values:
            e.fields = UnpicklerZoneInfo(io.BytesIO(e.fields)).load()
            e.result = UnpicklerZoneInfo(io.BytesIO(e.result)).load()
            e.start = UnpicklerZoneInfo(io.BytesIO(e.start)).load()
            e.not_competing = bool(e.not_competing)
        return values

    def get_entry_by_name(self, event_id: int, first_name: str, last_name: str):
        values = self.db.query(
            "SELECT "
            "entries.id,"
            "entries.event_id,"
            "competitors.id,"
            "competitors.first_name,"
            "competitors.last_name,"
            "competitors.gender,"
            "competitors.year,"
            "classes.id,"
            "classes.name,"
            "clubs.id,"
            "clubs.name,"
            "entries.not_competing,"
            "entries.chip,"
            "entries.fields,"
            "entries.result,"
            "entries.start "
            "FROM entries "
            "LEFT JOIN competitors ON entries.competitor_id = competitors.id "
            "LEFT JOIN classes ON entries.class_id = classes.id "
            "LEFT JOIN clubs ON entries.club_id = clubs.id "
            "WHERE entries.event_id = " + web.db.sqlquote(event_id) + " AND "
            "competitors.first_name = " + web.db.sqlquote(first_name) + " AND "
            "competitors.last_name = " + web.db.sqlquote(last_name) + ";"
        )
        values.names[values.names.index("id", 1)] = "competitor_id"
        values.names[values.names.index("id", 1)] = "class_id"
        values.names[values.names.index("id", 1)] = "club_id"
        values.names[values.names.index("name")] = "class_"
        values.names[values.names.index("name")] = "club"
        values = list(values)
        for e in values:
            e.fields = UnpicklerZoneInfo(io.BytesIO(e.fields)).load()
            e.result = UnpicklerZoneInfo(io.BytesIO(e.result)).load()
            e.start = UnpicklerZoneInfo(io.BytesIO(e.start)).load()
            e.not_competing = bool(e.not_competing)
        return values

    def add_entry(
        self,
        event_id: int,
        competitor_id: Optional[int],
        first_name: str,
        last_name: str,
        gender: str,
        year: Optional[int],
        class_id: int,
        club_id: Optional[int],
        not_competing: bool,
        chip: str,
        fields: Dict[int, str],
        status: result_type.ResultStatus,
        start_time: Optional[datetime.datetime],
    ) -> int:
        if self.get_event(id=event_id) == []:
            raise EventNotFoundError

        if competitor_id is None:
            for com in self.get_competitors():
                if com.first_name == first_name and com.last_name == last_name:
                    competitor_id = com.id
                    if gender == "":
                        gender = com.gender
                    if year is None:
                        year = com.year
                    if chip == "":
                        chip = com.chip
                    if club_id is None:
                        club_id = com.club_id
                    break
            else:
                competitor_id = self.add_competitor(
                    first_name=first_name,
                    last_name=last_name,
                    club_id=club_id,
                    gender=gender,
                    year=year,
                    chip=chip,
                )
        try:
            competitor = list(self.get_competitor(id=competitor_id))
            self.update_competitor(
                id=competitor[0].id,
                first_name=first_name,
                last_name=last_name,
                gender=gender,
                year=year,
                club_id=competitor[0].club_id
                if competitor[0].club_id is not None
                else club_id,
                chip=competitor[0].chip if competitor[0].chip != "" else chip,
            )
            return self.db.insert(
                "entries",
                event_id=event_id,
                competitor_id=competitor_id,
                class_id=class_id,
                club_id=club_id,
                not_competing=not_competing,
                result=pickle.dumps(result_type.PersonRaceResult(status=status)),
                start=pickle.dumps(start_type.PersonRaceStart(start_time=start_time)),
                chip=chip,
                fields=pickle.dumps(fields),
            )
        except sqlite3.IntegrityError:
            raise ConstraintError("Competitor already registered for this event")

    def add_entry_result(
        self,
        event_id: int,
        chip: str,
        start_time: Optional[datetime.datetime],
        result: result_type.PersonRaceResult,
    ) -> int:
        if self.get_event(id=event_id) == []:
            raise EventNotFoundError

        try:
            return self.db.insert(
                "entries",
                event_id=event_id,
                competitor_id=None,
                class_id=None,
                club_id=None,
                not_competing=False,
                result=pickle.dumps(result),
                start=pickle.dumps(start_type.PersonRaceStart(start_time=start_time)),
                chip=chip,
                fields=pickle.dumps({}),
            )
        except sqlite3.IntegrityError:
            raise ConstraintError("Competitor already registered for this event")

    def update_entry(
        self,
        id: int,
        first_name: str,
        last_name: str,
        gender: str,
        year: Optional[int],
        class_id: int,
        club_id: Optional[int],
        not_competing: bool,
        chip: str,
        fields: Dict[int, str],
        status: result_type.ResultStatus,
        start_time: Optional[datetime.datetime],
    ) -> None:
        if self.get_entry(id=id) == []:
            raise KeyError

        entry = self.get_entry(id=id)
        competitor = list(self.get_competitor(id=entry[0].competitor_id))
        self.update_competitor(
            id=competitor[0].id,
            first_name=first_name,
            last_name=last_name,
            gender=gender,
            year=year,
            club_id=competitor[0].club_id,
            chip=competitor[0].chip,
        )

        result = entry[0].result
        result.status = status
        start = entry[0].start
        start.start_time = start_time
        self.db.update(
            "entries",
            where="id=" + web.db.sqlquote(id),
            class_id=class_id,
            club_id=club_id,
            not_competing=not_competing,
            chip=chip,
            fields=pickle.dumps(fields),
            result=pickle.dumps(result),
            start=pickle.dumps(start),
        )

    def update_entry_result(
        self,
        id: int,
        chip: str,
        start_time: Optional[datetime.datetime],
        result: result_type.PersonRaceResult,
    ) -> None:
        if self.get_entry(id=id) == []:
            raise KeyError

        entry = self.get_entry(id=id)
        start = entry[0].start
        start.start_time = start_time
        self.db.update(
            "entries",
            where="id=" + web.db.sqlquote(id),
            chip=chip,
            result=pickle.dumps(result),
            start=pickle.dumps(start),
        )

    def delete_entries(self, event_id: int):
        self.db.delete("entries", where="event_id=" + web.db.sqlquote(event_id))

    def delete_entry(self, id: int):
        self.db.delete("entries", where="id=" + web.db.sqlquote(id))

    def import_entries(self, event_id: int, entries):
        if self.get_event(id=event_id) == []:
            raise EventNotFoundError

        list_of_entries = []
        for c in entries:
            class_id = None
            class_ = None
            for cla in self.get_classes(event_id=event_id):
                if cla.name == c["class_"]:
                    class_id = cla.id
                    class_ = cla
                    break
            else:
                class_id = self.add_class(
                    event_id=event_id,
                    name=c["class_"],
                    short_name=None,
                    course_id=None,
                    params=ClassParams(),
                )

            club_id = None
            if c["club"] != "":
                for clb in self.get_clubs():
                    if clb.name == c["club"]:
                        club_id = clb.id
                        break
                else:
                    club_id = self.add_club(c["club"])

            gender = c["gender"] if "gender" in c else ""
            year = c["year"] if "year" in c else None
            competitor = self.get_competitor_by_name(
                first_name=c["first_name"], last_name=c["last_name"]
            ).first()
            if competitor is None:
                competitor_id = self.add_competitor(
                    first_name=c["first_name"],
                    last_name=c["last_name"],
                    club_id=club_id,
                    gender=gender,
                    year=year,
                    chip=c["chip"] if "chip" in c else "",
                )
            else:
                competitor_id = competitor.id
                # update gender and year in competitor
                gender = gender if gender != "" else competitor.gender
                year = year if year is not None else competitor.year
                if gender != competitor.gender or year != competitor.year:
                    self.update_competitor(
                        id=competitor.id,
                        first_name=competitor.first_name,
                        last_name=competitor.last_name,
                        club_id=competitor.club_id,
                        gender=gender,
                        year=year,
                        chip=competitor.chip,
                    )

            # update result
            if c["result"].has_punches():
                try:
                    course_id = class_["course_id"]
                    class_params = class_["params"]
                    controls = self.get_course(id=course_id)[0]["controls"]
                except:
                    class_params = ClassParams()
                    controls = []
                c["result"].compute_result(
                    controls=controls,
                    class_params=class_params,
                    start_time=c["result"].start_time,
                    year=year,
                    gender=gender if gender != "" else None,
                )

            entry = self.get_entry_by_name(
                event_id=event_id, first_name=c["first_name"], last_name=c["last_name"]
            )
            if entry == []:
                if "result" in c:
                    result = copy.deepcopy(c["result"])
                else:
                    result = result_type.PersonRaceResult()
                if "start" in c:
                    start = copy.deepcopy(c["start"])
                else:
                    start = start_type.PersonRaceStart()
                list_of_entries.append(
                    {
                        "event_id": event_id,
                        "competitor_id": competitor_id,
                        "class_id": class_id,
                        "club_id": club_id,
                        "not_competing": c["not_competing"]
                        if "not_competing" in c
                        else False,
                        "chip": c["chip"] if "chip" in c else "",
                        "fields": pickle.dumps(c["fields"] if "fields" in c else {}),
                        "result": pickle.dumps(result),
                        "start": pickle.dumps(start),
                    }
                )
            else:
                fields = entry[0].fields
                if "fields" in c:
                    fields = copy.deepcopy(c["fields"])
                result = entry[0].result
                if "result" in c:
                    result = copy.deepcopy(c["result"])
                start = entry[0].start
                if "start" in c:
                    start = copy.deepcopy(c["start"])
                self.db.update(
                    "entries",
                    where="id=" + web.db.sqlquote(entry[0].id),
                    class_id=class_id,
                    club_id=club_id,
                    not_competing=c["not_competing"]
                    if "not_competing" in c
                    else entry[0].not_competing,
                    chip=c["chip"] if "chip" in c else entry[0].chip,
                    fields=pickle.dumps(fields),
                    result=pickle.dumps(result),
                    start=pickle.dumps(start),
                )

        self.db.supports_multiple_insert = True
        for i in range(0, len(entries), 25):
            self.db.multiple_insert(
                "entries", list_of_entries[i : min(i + 25, len(entries))]
            )

    def get_events(self) -> List:
        values = self.db.query(
            "SELECT "
            "events.id,"
            "events.name,"
            "events.date,"
            "events.key,"
            "events.publish,"
            "events.series,"
            "events.fields "
            "FROM events "
            "ORDER BY events.name ASC;"
        )
        values = list(values)
        for e in values:
            e.date = datetime.datetime.strptime(e.date, "%Y-%m-%d").date()
            e.fields = UnpicklerZoneInfo(io.BytesIO(e.fields)).load()
            e.publish = bool(e.publish)
        return values

    def get_event(self, id: int) -> List:
        values = self.db.where("events", id=id)
        values = list(values)
        for e in values:
            e.date = datetime.datetime.strptime(e.date, "%Y-%m-%d").date()
            e.fields = UnpicklerZoneInfo(io.BytesIO(e.fields)).load()
            e.publish = bool(e.publish)
        return values

    def add_event(
        self,
        name: str,
        date: datetime.date,
        key: Optional[str],
        publish: bool,
        series: Optional[str],
        fields: List[str] = [],
    ) -> int:
        try:
            return self.db.insert(
                "events",
                name=name,
                date=date.isoformat(),
                key=key,
                publish=publish,
                series=series,
                fields=pickle.dumps(fields),
            )

        except sqlite3.IntegrityError:
            raise ConstraintError("Event or event key already exist")

    def update_event(
        self,
        id: int,
        name: str,
        date: datetime.date,
        key: Optional[str],
        publish: bool,
        series: Optional[str],
        fields: List[str] = [],
    ) -> None:
        try:
            nr_of_rows = self.db.update(
                "events",
                where="id=" + web.db.sqlquote(id),
                name=name,
                date=date.isoformat(),
                key=key,
                publish=publish,
                series=series,
                fields=pickle.dumps(fields),
            )
            if nr_of_rows == 0:
                raise KeyError
        except sqlite3.IntegrityError:
            raise ConstraintError("Event or event key already exist")

    def delete_event(self, id: int) -> None:
        self.db.delete("events", where="id=" + web.db.sqlquote(id))

    def get_series_settings(self) -> series_type.Settings:
        settings = series_type.Settings()
        values = self.db.query(
            "SELECT name, nr_of_best_results, mode, maximum_points, decimal_places FROM settings;"
        )
        values = list(values)
        if values != []:
            settings.name = values[0].name
            settings.nr_of_best_results = values[0].nr_of_best_results
            settings.mode = values[0].mode
            settings.maximum_points = values[0].maximum_points
            settings.decimal_places = values[0].decimal_places
        return settings

    def update_series_settings(self, settings: series_type.Settings) -> None:
        nr_of_rows = self.db.update(
            tables="settings",
            where="True",
            name=settings.name,
            nr_of_best_results=settings.nr_of_best_results,
            mode=settings.mode,
            maximum_points=settings.maximum_points,
            decimal_places=settings.decimal_places,
        )
        if nr_of_rows == 0:
            self.db.insert(
                tablename="settings",
                name=settings.name,
                nr_of_best_results=settings.nr_of_best_results,
                mode=settings.mode,
                maximum_points=settings.maximum_points,
                decimal_places=settings.decimal_places,
            )



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from typing import Dict
from typing import Optional
from typing import Any

from ooresults.repo.result_type import ResultStatus
from ooresults.pdf.pdf import PDF


def create_pdf(
    event: Dict, results: Dict, include_dns: bool = False, landscape: bool = True
) -> bytes:
    W_SPACE = 4
    W_RANK = 9
    W_NAME = 60
    W_YEAR = 10
    W_CLUB = 60
    W_RUNTIME = 15
    W_VALUE = 20
    W_TOTAL = 24

    if landscape:
        W_CLUB = 175
    else:
        W_CLUB = 175 - 87

    # Instantiation of inherited class
    pdf = PDF(name=event.name, landscape=landscape)
    pdf.set_margin(margin=10)
    pdf.set_auto_page_break(auto=True, margin=20)
    pdf.add_page()

    def cell(w: int, h: Optional[int] = None, txt: str = "", align: str = "L") -> None:
        while pdf.get_string_width(txt) > w:
            txt = txt[:-1]
        pdf.cell(w=w, h=h, txt=txt, align=align)

    def print_time(width: int, time: int, status: ResultStatus) -> None:
        txt = ""
        if status == ResultStatus.OK:
            txt = "{:d}:{:02d}".format(time // 60, time % 60)
        cell(w=width, h=None, txt=txt, align="R")

    def print_time_or_status(width: int, time: int, status: ResultStatus) -> None:
        txt = pdf.MAP_STATUS[status]
        if status == ResultStatus.OK:
            txt = "{:d}:{:02d}".format(time // 60, time % 60)
        cell(w=width, h=None, txt=txt, align="R")

    def print_points(width: int, points: Optional[float], status: ResultStatus) -> None:
        txt = ""
        if points is not None and status == ResultStatus.OK:
            txt = "{:.2f}".format(points)
        cell(w=width, h=None, txt=txt, align="R")

    def print_points_or_status(
        width: int, points: Optional[float], status: ResultStatus
    ) -> None:
        txt = pdf.MAP_STATUS[status]
        if points is not None and status == ResultStatus.OK:
            txt = "{:.2f}".format(points)
        cell(w=width, h=None, txt=txt, align="R")

    for i, class_results in enumerate(results):
        class_, ranked_results = class_results
        if i > 0:
            # insert a page break if there is not enough space left on the
            # page for the class header, the table header and two table rows
            if pdf.will_page_break(height=33):
                pdf.add_page()
            else:
                pdf.ln()
                pdf.ln()

        width = 0
        if class_.params.otype == "score":
            width += W_RUNTIME
            width += W_VALUE
            width += W_VALUE
            if class_.params.apply_handicap_rule:
                width += W_VALUE
            width += W_TOTAL
        else:
            if (
                class_.params.penalty_controls is not None
                or class_.params.penalty_overtime is not None
                or class_.params.apply_handicap_rule
            ):
                width += W_RUNTIME
            if class_.params.penalty_controls is not None:
                width += W_VALUE
            if class_.params.penalty_overtime is not None:
                width += W_VALUE
            if class_.params.apply_handicap_rule:
                width += W_VALUE
            width += W_TOTAL

        # number of entries
        nr_entries = 0
        for result in ranked_results:
            if include_dns or result["result"].status not in (
                ResultStatus.INACTIVE,
                ResultStatus.DID_NOT_START,
            ):
                nr_entries += 1

        pdf.set_font(family="Carlito", style="B", size=12)
        pdf.cell(txt=f"{class_.name}   ({nr_entries})")

        # print possible voided legs
        if ranked_results and ranked_results[0]["result"] is not None:
            voided_legs = ranked_results[0]["result"].voided_legs()
            if voided_legs:
                pdf.cell(txt=f'(Voided legs: {", ".join(voided_legs)})')

        pdf.ln()
        pdf.ln()
        pdf.set_font(family="Carlito", style="I", size=10)
        cell(w=W_RANK, h=None, txt="Pl", align="R")
        cell(w=W_SPACE, h=None, txt="")
        cell(w=W_NAME, h=None, txt="Name", align="L")
        cell(w=W_SPACE, h=None, txt="")
        cell(w=W_YEAR, h=None, txt="Jg", align="R")
        cell(w=W_SPACE, h=None, txt="")
        cell(w=W_CLUB - width, h=None, txt="Verein", align="L")
        cell(w=W_SPACE, h=None, txt="")
        if class_.params.otype == "score":
            if class_.params.apply_handicap_rule:
                cell(w=W_VALUE, h=None, txt="Faktor", align="R")
            cell(w=W_RUNTIME, h=None, txt="Laufzeit", align="R")
            cell(w=W_VALUE, h=None, txt="Sc-Posten", align="R")
            cell(w=W_VALUE, h=None, txt="Sc-Zeit", align="R")
            cell(w=W_TOTAL, h=None, txt="Score", align="R")
        else:
            if (
                class_.params.penalty_controls is not None
                or class_.params.penalty_overtime is not None
                or class_.params.apply_handicap_rule
            ):
                cell(w=W_RUNTIME, h=None, txt="Laufzeit", align="R")
            if class_.params.penalty_controls is not None:
                cell(w=W_VALUE, h=None, txt="Str-Posten", align="R")
            if class_.params.penalty_overtime is not None:
                cell(w=W_VALUE, h=None, txt="Str-Zeit", align="R")
            if class_.params.apply_handicap_rule:
                cell(w=W_VALUE, h=None, txt="Faktor", align="R")
            cell(w=W_TOTAL, h=None, txt="Zeit", align="R")
        pdf.ln()
        pdf.ln()

        ranked = False
        for result in ranked_results:

            def get(d: Dict, key: str) -> Any:
                s = d.get(key, "")
                return s if s is not None else ""

            if include_dns or result["result"].status not in (
                ResultStatus.INACTIVE,
                ResultStatus.DID_NOT_START,
            ):
                pdf.set_font(family="Carlito", size=12)
                if result.rank is not None:
                    ranked = True
                elif ranked:
                    ranked = False
                    pdf.ln()
                cell(
                    w=W_RANK,
                    h=None,
                    txt=str(get(result, "rank"))
                    if not get(result, "not_competing")
                    else "AK",
                    align="R",
                )
                cell(w=W_SPACE, h=None, txt="")
                cell(
                    w=W_NAME,
                    h=None,
                    txt=get(result, "last_name") + " " + get(result, "first_name"),
                    align="L",
                )
                cell(w=W_SPACE, h=None, txt="")
                cell(w=W_YEAR, h=None, txt=str(get(result, "year")), align="R")
                cell(w=W_SPACE, h=None, txt="")
                cell(w=W_CLUB - width, h=None, txt=get(result, "club"), align="L")
                cell(w=W_SPACE, h=None, txt="")
                status = result["result"].status
                if class_.params.otype == "score":
                    if class_.params.apply_handicap_rule:
                        if status == ResultStatus.OK:
                            cell(
                                w=W_VALUE,
                                h=None,
                                txt="{:1.4f}".format(
                                    result["result"].extensions.get("factor", 1)
                                ),
                                align="R",
                            )
                        else:
                            cell(w=W_VALUE, h=None, txt="")
                    print_time(
                        width=W_RUNTIME, time=result["result"].time, status=status
                    )
                    print_points(
                        width=W_VALUE,
                        points=result["result"].extensions.get("score_controls", None),
                        status=status,
                    )
                    print_points(
                        width=W_VALUE,
                        points=result["result"].extensions.get("score_overtime", None),
                        status=status,
                    )
                    print_points_or_status(
                        width=W_TOTAL,
                        points=result["result"].extensions.get("score", None),
                        status=status,
                    )
                else:
                    if (
                        class_.params.penalty_controls is not None
                        or class_.params.penalty_overtime is not None
                        or class_.params.apply_handicap_rule
                    ):
                        print_time(
                            width=W_RUNTIME,
                            time=result["result"].extensions.get("running_time", None),
                            status=status,
                        )
                    if class_.params.penalty_controls is not None:
                        print_time(
                            width=W_VALUE,
                            time=result["result"].extensions.get(
                                "penalties_controls", None
                            ),
                            status=status,
                        )
                    if class_.params.penalty_overtime is not None:
                        print_time(
                            width=W_VALUE,
                            time=result["result"].extensions.get(
                                "penalties_overtime", None
                            ),
                            status=status,
                        )
                    if class_.params.apply_handicap_rule:
                        if status == ResultStatus.OK:
                            cell(
                                w=W_VALUE,
                                h=None,
                                txt="{:1.4f}".format(
                                    result["result"].extensions.get("factor", 1)
                                ),
                                align="R",
                            )
                        else:
                            cell(w=W_VALUE, h=None, txt="")
                    print_time_or_status(
                        width=W_TOTAL, time=result["result"].time, status=status
                    )
                pdf.ln()

    return bytes(pdf.output())



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import logging
import copy
import pathlib
from typing import List
from typing import Dict
from typing import Tuple

import web

from ooresults.handler import model
from ooresults.repo.result_type import ResultStatus
import ooresults.pdf.result
import ooresults.pdf.splittimes
from ooresults.utils.globals import t_globals


templates = pathlib.Path(__file__).resolve().parent.parent / "templates"
render = web.template.render(templates, globals=t_globals)


def build_results(classes: List[Dict], results: List[Dict]):
    all_results = []
    for class_ in classes:
        if class_.name == "Organizer":
            continue
        # filter results belonging to this class
        class_results = []
        for r in results:
            if r["class_"] == class_.name:
                class_results.append(r)

        if class_results != []:
            # set the time column used for sorting
            def result_equal(c1, c2) -> bool:
                if class_.params.otype == "score":
                    return (
                        c1.extensions["score"] == c2.extensions["score"]
                        and c1.time == c2.time
                    )
                else:
                    return c1.time == c2.time

            # sort the results
            if class_.params.otype == "score":
                class_results.sort(key=lambda c: c["last_name"] + "," + c["first_name"])
                class_results.sort(
                    key=lambda c: c["result"].time
                    if c["result"].time is not None
                    else 0
                )
                class_results.sort(
                    key=lambda c: c["result"].extensions["score"]
                    if c["result"].extensions.get("score", None) is not None
                    else -999999,
                    reverse=True,
                )
            else:
                class_results.sort(key=lambda c: c["last_name"] + "," + c["first_name"])
                class_results.sort(
                    key=lambda c: c["result"].time
                    if c["result"].time is not None
                    else 0
                )

            def sort_for_rank(e) -> int:
                mapping = {
                    (ResultStatus.OK, False): 0,
                    (ResultStatus.MISSING_PUNCH, False): 1,
                    (ResultStatus.OVER_TIME, False): 2,
                    (ResultStatus.DID_NOT_FINISH, False): 3,
                    (ResultStatus.DISQUALIFIED, False): 4,
                    (ResultStatus.OK, True): 5,
                    (ResultStatus.MISSING_PUNCH, True): 6,
                    (ResultStatus.OVER_TIME, True): 7,
                    (ResultStatus.DID_NOT_FINISH, True): 8,
                    (ResultStatus.DISQUALIFIED, True): 9,
                    (ResultStatus.DID_NOT_START, False): 10,
                    (ResultStatus.DID_NOT_START, True): 11,
                }
                return mapping.get((e["result"].status, e["not_competing"]), 99)

            class_results.sort(key=sort_for_rank)

            # add rank: Optional[int]
            for i, r in enumerate(class_results):
                if r["result"].status == ResultStatus.OK and not r["not_competing"]:
                    if i > 0 and result_equal(
                        class_results[i]["result"], class_results[i - 1]["result"]
                    ):
                        r["rank"] = class_results[i - 1]["rank"]
                    else:
                        r["rank"] = i + 1
                else:
                    r["rank"] = None

            # add points
            if class_.params.otype != "score":
                ref_time = class_results[0]["result"].time
                for r in class_results:
                    if r["rank"] is not None:
                        r["points"] = ref_time / r["result"].time
                    elif r["result"].status == ResultStatus.OK and r["not_competing"]:
                        r["points"] = 0
                    elif r["result"].status in (
                        ResultStatus.MISSING_PUNCH,
                        ResultStatus.DID_NOT_FINISH,
                        ResultStatus.OVER_TIME,
                        ResultStatus.DISQUALIFIED,
                    ):
                        r["points"] = 0

            all_results.append((class_, class_results))
    return all_results


def build_columns(class_results: List[Tuple[Dict, Dict]]) -> set:
    columns = set()
    for class_, _ in class_results:
        if class_.params.apply_handicap_rule:
            columns.add("factor")
        if class_.params.penalty_controls is not None:
            columns.add("penalties_controls")
        if class_.params.penalty_overtime is not None:
            columns.add("penalties_overtime")
        if class_.params.otype == "score":
            columns.add("score")
    return columns


class Update:
    def POST(self):
        """Update data"""
        data = web.input()
        event_id = int(data.event_id) if data.event_id != "" else -1
        event = list(model.get_event(id=event_id))
        event = event[0] if event != [] else {}

        classes = list(model.get_classes(event_id=event_id))
        entry_list = list(model.get_entries(event_id=event_id))
        class_results = build_results(
            classes=classes, results=copy.deepcopy(entry_list)
        )
        columns = build_columns(class_results)
        return render.results_table(event, class_results, columns)


class PdfResult:
    def POST(self):
        """Print results"""
        data = web.input()
        event_id = int(data.event_id) if data.event_id != "" else -1
        include_dns = "res_include_dns" in data

        try:
            event = list(model.get_event(id=event_id))
            event = event[0] if event != [] else {}
            classes = list(model.get_classes(event_id=event_id))

            entry_list = list(model.get_entries(event_id=event_id))
            class_results = build_results(classes, copy.deepcopy(entry_list))
            columns = build_columns(class_results)
            content = ooresults.pdf.result.create_pdf(
                event=event,
                results=class_results,
                include_dns=include_dns,
                landscape=len(columns) > 0,
            )
            return content

        except KeyError:
            raise web.conflict("Entry deleted")
        except:
            logging.exception("Internal server error")
            raise

        return content


class PdfSplittimes:
    def POST(self):
        """Print results"""
        data = web.input()
        event_id = int(data.event_id) if data.event_id != "" else -1
        landscape = "res_landscape" in data

        try:
            event = list(model.get_event(id=event_id))
            event = event[0] if event != [] else {}
            classes = list(model.get_classes(event_id=event_id))

            entry_list = list(model.get_entries(event_id=event_id))
            class_results = build_results(classes, copy.deepcopy(entry_list))
            content = ooresults.pdf.splittimes.create_pdf(
                event=event, results=class_results, landscape=landscape
            )
            return content

        except KeyError:
            raise web.conflict("Entry deleted")
        except:
            logging.exception("Internal server error")
            raise

        return content



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import json
import pathlib
import tempfile

from ooresults import user


def test_default_values_if_file_not_exist():
    with tempfile.TemporaryDirectory() as td:
        path = pathlib.Path(td) / "users.json"

        user.Users.update(path=path)
        assert path.exists()

        with open(path) as f:
            data = json.load(f)
        assert data == [{"username": "admin", "password": "admin"}]
        assert user.Users.path == path
        assert user.Users.users == {"admin": "admin"}


def test_update_values_if_file_exist():
    users = [
        {"username": "a", "password": "x"},
        {"username": "b", "password": "y"},
        {"username": "c", "password": "z"},
    ]
    with tempfile.TemporaryDirectory() as td:
        path = pathlib.Path(td) / "users.json"
        with open(path, "w") as f:
            f.write(json.dumps(users))

        user.Users.update(path=path)
        assert path.exists()

        with open(path) as f:
            data = json.load(f)
        assert data == users
        assert user.Users.path == path
        assert user.Users.users == {"a": "x", "b": "y", "c": "z"}


def test_check_username_and_password():
    user.Users.users = {
        "abc": "p1",
        "def": "p2",
        "ghi": "p3",
    }
    assert user.Users.check("abc", "p1")
    assert user.Users.check("def", "p2")
    assert user.Users.check("ghi", "p3")
    assert not user.Users.check("abc", "p2")
    assert not user.Users.check("abc", "p3")
    assert not user.Users.check("xxx", "pp")



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from ooresults.repo import result_type


def parse(content: bytes):
    try:
        content = content.decode(encoding="utf-8")
    except:
        content = content.decode(encoding="windows-1252")

    # Format:
    #
    # Bahn A - Lang
    #  1  Christian Lindner                            OC Gelb                       1:07:58
    #  2  Gisela Schmidt                               Individuals/No club           1:08:02
    #     Manuela Bodental                             Individuals/No club               mp
    #
    # Bahn B - Mittel
    #  1  Hugo Peter Wolf                              Individuals/No club             28:29
    #  1  Olaf Scholz                                  OC Rot                          28:29

    results = []
    for line in content.split("\n"):
        if line != "":
            if line[0] != " ":
                class_ = line.strip()
            else:
                r = {}
                r["class_"] = class_

                a, _, b = line[4:].strip().partition("  ")
                r["first_name"], _, r["last_name"] = a.rpartition(" ")
                r["club"], _, b = b.strip().partition("  ")
                h_m_s, _, b = b.strip().partition("  ")

                MAP_STATUS = {
                    "OK": result_type.ResultStatus.OK,
                    "MP": result_type.ResultStatus.MISSING_PUNCH,
                    "DNS": result_type.ResultStatus.DID_NOT_START,
                    "DNF": result_type.ResultStatus.DID_NOT_FINISH,
                    "DISQ": result_type.ResultStatus.DISQUALIFIED,
                }
                if h_m_s.upper() in ["MP", "DNS", "DNF", "DISQ"]:
                    r["result"] = result_type.PersonRaceResult(
                        status=MAP_STATUS[h_m_s.upper()], time=None
                    )
                else:
                    t = 0
                    for i in h_m_s.split(":"):
                        t = 60 * t + int(i)
                    r["result"] = result_type.PersonRaceResult(
                        status=result_type.ResultStatus.OK, time=t
                    )

                results.append(r)

    return results



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from decimal import Decimal

import pytest

from ooresults.handler import series
from ooresults.repo import result_type
from ooresults.repo import series_type


class Event:
    def __init__(self, name: str, date: str, series: bool):
        self.name = name
        self.date = date
        self.series = series


class Class_:
    def __init__(self, name: str):
        self.name = name


@pytest.fixture
def event_1():
    return Event(name="ev1", date="2021-03-20", series="Run 1")


@pytest.fixture
def event_2():
    return Event(name="ev2", date="2021-02-20", series=None)


@pytest.fixture
def event_3():
    return Event(name="ev3", date="2021-02-21", series="Run 3")


@pytest.fixture
def event_4():
    return Event(name="ev4", date="2021-02-21", series="Run 4")


@pytest.fixture
def event_5():
    return Event(name="ev5", date="2021-02-19", series="Run 5")


def test_use_only_events_in_series(event_1, event_2, event_3):
    events = [event_1, event_2, event_3]
    data = series.create_event_list(events=events)
    assert data == [event_3, event_1]


def test_sort_events_by_date(event_1, event_3, event_5):
    events = [event_1, event_3, event_5]
    data = series.create_event_list(events=events)
    assert data == [event_5, event_3, event_1]

    events = [event_1, event_3, event_5]
    data = series.create_event_list(events=events)
    assert data == [event_5, event_3, event_1]


def test_sort_events_by_name_if_dates_are_equal(event_3, event_4):
    events = [event_3, event_4]
    data = series.create_event_list(events=events)
    assert data == [event_3, event_4]

    events = [event_4, event_3]
    data = series.create_event_list(events=events)
    assert data == [event_3, event_4]


def test_no_results():
    class_a = Class_(name="Bahn A - Lang")
    data = series.build_total_results(
        settings=series_type.Settings(),
        list_of_results=[
            [
                (
                    class_a,
                    [
                        {
                            "first_name": "Angela",
                            "last_name": "Merkel",
                            "gender": "",
                            "year": "",
                            "class_": "Bahn A - Lang",
                            "result": result_type.PersonRaceResult(),
                        },
                        {
                            "first_name": "Claudia",
                            "last_name": "Merkel",
                            "gender": "",
                            "year": "",
                            "class_": "Bahn A - Lang",
                            "result": result_type.PersonRaceResult(),
                        },
                        {
                            "first_name": "Birgit",
                            "last_name": "Merkel",
                            "gender": "",
                            "year": "",
                            "class_": "Bahn A - Lang",
                            "result": result_type.PersonRaceResult(),
                        },
                    ],
                ),
            ]
        ],
    )
    assert data == [(class_a.name, [])]


def test_best_3_of_1_race():
    class_a = Class_(name="Bahn A - Lang")
    data = series.build_total_results(
        settings=series_type.Settings(
            nr_of_best_results=3,
        ),
        list_of_results=[
            [
                (
                    class_a,
                    [
                        {
                            "first_name": "Angela",
                            "last_name": "Merkel",
                            "gender": "",
                            "year": "",
                            "class_": "Bahn A - Lang",
                            "result": result_type.PersonRaceResult(),
                            "points": 0.365,
                        },
                    ],
                ),
            ],
        ],
    )
    print(data)
    assert data == [
        (
            class_a.name,
            [
                {
                    "first_name": "Angela",
                    "last_name": "Merkel",
                    "year": "",
                    "club": None,
                    "races": {0: Decimal("36.5")},
                    "organizer": {},
                    "sum": Decimal("36.5"),
                    "rank": 1,
                },
            ],
        ),
    ]


def test_best_3_of_2_races():
    class_a = Class_(name="Bahn A - Lang")
    data = series.build_total_results(
        settings=series_type.Settings(
            nr_of_best_results=3,
        ),
        list_of_results=[
            [
                (
                    class_a,
                    [
                        {
                            "first_name": "Angela",
                            "last_name": "Merkel",
                            "gender": "",
                            "year": "",
                            "class_": "Bahn A - Lang",
                            "result": result_type.PersonRaceResult(),
                            "points": 0.365,
                        },
                    ],
                ),
            ],
            [
                (
                    class_a,
                    [
                        {
                            "first_name": "Angela",
                            "last_name": "Merkel",
                            "gender": "",
                            "year": "",
                            "class_": "Bahn A - Lang",
                            "result": result_type.PersonRaceResult(),
                            "points": 0.991,
                        },
                    ],
                ),
            ],
        ],
    )
    print(data)
    assert data == [
        (
            class_a.name,
            [
                {
                    "first_name": "Angela",
                    "last_name": "Merkel",
                    "year": "",
                    "club": None,
                    "races": {0: Decimal("36.5"), 1: Decimal("99.1")},
                    "organizer": {},
                    "sum": Decimal("135.6"),
                    "rank": 1,
                },
            ],
        ),
    ]


def test_best_3_of_4_races():
    class_a = Class_(name="Bahn A - Lang")
    data = series.build_total_results(
        settings=series_type.Settings(
            nr_of_best_results=3,
        ),
        list_of_results=[
            [
                (
                    class_a,
                    [
                        {
                            "first_name": "Angela",
                            "last_name": "Merkel",
                            "gender": "",
                            "year": "",
                            "class_": "Bahn A - Lang",
                            "result": result_type.PersonRaceResult(),
                            "points": 0.365,
                        },
                    ],
                ),
            ],
            [
                (
                    class_a,
                    [
                        {
                            "first_name": "Angela",
                            "last_name": "Merkel",
                            "gender": "",
                            "year": "",
                            "class_": "Bahn A - Lang",
                            "result": result_type.PersonRaceResult(),
                            "points": 0.991,
                        },
                    ],
                ),
            ],
            [
                (
                    class_a,
                    [
                        {
                            "first_name": "Angela",
                            "last_name": "Merkel",
                            "gender": "",
                            "year": "",
                            "class_": "Bahn A - Lang",
                            "result": result_type.PersonRaceResult(),
                            "points": 0.012,
                        },
                    ],
                ),
            ],
            [
                (
                    class_a,
                    [
                        {
                            "first_name": "Angela",
                            "last_name": "Merkel",
                            "gender": "",
                            "year": "",
                            "class_": "Bahn A - Lang",
                            "result": result_type.PersonRaceResult(),
                            "points": 0.2,
                        },
                    ],
                ),
            ],
        ],
    )
    print(data)
    assert data == [
        (
            class_a.name,
            [
                {
                    "first_name": "Angela",
                    "last_name": "Merkel",
                    "year": "",
                    "club": None,
                    "races": {
                        0: Decimal("36.5"),
                        1: Decimal("99.1"),
                        2: Decimal("1.2"),
                        3: Decimal("20.0"),
                    },
                    "organizer": {},
                    "sum": Decimal("155.6"),
                    "rank": 1,
                },
            ],
        ),
    ]


def test_bonus_1_race():
    class_a = Class_(name="Bahn A - Lang")
    data = series.build_total_results(
        settings=series_type.Settings(),
        list_of_results=[
            [
                (
                    class_a,
                    [
                        {
                            "first_name": "Angela",
                            "last_name": "Merkel",
                            "gender": "",
                            "year": "",
                            "class_": "Bahn A - Lang",
                            "result": result_type.PersonRaceResult(),
                            "points": 0.365,
                        },
                    ],
                ),
            ],
            [
                (class_a, []),
            ],
        ],
        organizers=[
            [],
            [{"first_name": "Angela", "last_name": "Merkel"}],
        ],
    )
    print(data)
    assert data == [
        (
            class_a.name,
            [
                {
                    "first_name": "Angela",
                    "last_name": "Merkel",
                    "year": "",
                    "club": None,
                    "races": {0: Decimal("36.5")},
                    "organizer": {1: Decimal("18.25")},
                    "sum": Decimal("54.75"),
                    "rank": 1,
                },
            ],
        ),
    ]


def test_bonus_2_races():
    class_a = Class_(name="Bahn A - Lang")
    data = series.build_total_results(
        settings=series_type.Settings(),
        list_of_results=[
            [
                (
                    class_a,
                    [
                        {
                            "first_name": "Angela",
                            "last_name": "Merkel",
                            "gender": "",
                            "year": "",
                            "class_": "Bahn A - Lang",
                            "result": result_type.PersonRaceResult(),
                            "points": 0.365,
                        },
                    ],
                ),
            ],
            [
                (class_a, []),
            ],
            [
                (
                    class_a,
                    [
                        {
                            "first_name": "Angela",
                            "last_name": "Merkel",
                            "gender": "",
                            "year": "",
                            "class_": "Bahn A - Lang",
                            "result": result_type.PersonRaceResult(),
                            "points": 0.991,
                        },
                    ],
                ),
            ],
        ],
        organizers=[
            [],
            [{"first_name": "Angela", "last_name": "Merkel"}],
            [],
        ],
    )
    print(data)
    assert data == [
        (
            class_a.name,
            [
                {
                    "first_name": "Angela",
                    "last_name": "Merkel",
                    "year": "",
                    "club": None,
                    "races": {0: Decimal("36.5"), 2: Decimal("99.1")},
                    "organizer": {1: Decimal("67.8")},
                    "sum": Decimal("203.4"),
                    "rank": 1,
                },
            ],
        ),
    ]


def test_bonus_4_races():
    class_a = Class_(name="Bahn A - Lang")
    data = series.build_total_results(
        settings=series_type.Settings(
            nr_of_best_results=3,
        ),
        list_of_results=[
            [
                (
                    class_a,
                    [
                        {
                            "first_name": "Angela",
                            "last_name": "Merkel",
                            "year": "",
                            "gender": "",
                            "class_": "Bahn A - Lang",
                            "result": result_type.PersonRaceResult(),
                            "points": 0.365,
                        },
                    ],
                ),
            ],
            [
                (
                    class_a,
                    [
                        {
                            "first_name": "Angela",
                            "last_name": "Merkel",
                            "gender": "",
                            "year": "",
                            "class_": "Bahn A - Lang",
                            "result": result_type.PersonRaceResult(),
                            "points": 0.991,
                        },
                    ],
                ),
            ],
            [
                (class_a, []),
            ],
            [
                (
                    class_a,
                    [
                        {
                            "first_name": "Angela",
                            "last_name": "Merkel",
                            "gender": "",
                            "year": "",
                            "class_": "Bahn A - Lang",
                            "result": result_type.PersonRaceResult(),
                            "points": 0.012,
                        },
                    ],
                ),
            ],
            [
                (
                    class_a,
                    [
                        {
                            "first_name": "Angela",
                            "last_name": "Merkel",
                            "gender": "",
                            "year": "",
                            "class_": "Bahn A - Lang",
                            "result": result_type.PersonRaceResult(),
                            "points": 0.2,
                        },
                    ],
                ),
            ],
        ],
        organizers=[
            [],
            [],
            [{"first_name": "Angela", "last_name": "Merkel"}],
            [],
            [],
        ],
    )
    print(data)
    assert data == [
        (
            class_a.name,
            [
                {
                    "first_name": "Angela",
                    "last_name": "Merkel",
                    "year": "",
                    "club": None,
                    "races": {
                        0: Decimal("36.5"),
                        1: Decimal("99.1"),
                        3: Decimal("1.2"),
                        4: Decimal("20.0"),
                    },
                    "organizer": {2: Decimal("67.8")},
                    "sum": Decimal("203.4"),
                    "rank": 1,
                },
            ],
        ),
    ]


def test_ranking_1():
    class_a = Class_(name="Bahn A - Lang")
    data = series.build_total_results(
        settings=series_type.Settings(),
        list_of_results=[
            [
                (
                    class_a,
                    [
                        {
                            "first_name": "Angela",
                            "last_name": "Merkel",
                            "gender": "",
                            "year": "",
                            "class_": "Bahn A - Lang",
                            "result": result_type.PersonRaceResult(),
                            "points": 0.365,
                        },
                        {
                            "first_name": "Gerd",
                            "last_name": "Müller",
                            "gender": "",
                            "year": "",
                            "class_": "Bahn A - Lang",
                            "result": result_type.PersonRaceResult(),
                            "points": 0.991,
                        },
                    ],
                ),
            ],
        ],
    )
    print(data)
    assert data == [
        (
            class_a.name,
            [
                {
                    "first_name": "Gerd",
                    "last_name": "Müller",
                    "year": "",
                    "club": None,
                    "races": {0: Decimal("99.1")},
                    "organizer": {},
                    "sum": Decimal("99.1"),
                    "rank": 1,
                },
                {
                    "first_name": "Angela",
                    "last_name": "Merkel",
                    "year": "",
                    "club": None,
                    "races": {0: Decimal("36.5")},
                    "organizer": {},
                    "sum": Decimal("36.5"),
                    "rank": 2,
                },
            ],
        ),
    ]


def test_ranking_2():
    class_a = Class_(name="Bahn A - Lang")
    data = series.build_total_results(
        settings=series_type.Settings(),
        list_of_results=[
            [
                (
                    class_a,
                    [
                        {
                            "first_name": "Angela",
                            "last_name": "Merkel",
                            "gender": "",
                            "year": "",
                            "class_": "Bahn A - Lang",
                            "result": result_type.PersonRaceResult(),
                            "points": 0.995,
                        },
                    ],
                )
            ],
            [
                (
                    class_a,
                    [
                        {
                            "first_name": "Gerd",
                            "last_name": "Müller",
                            "gender": "",
                            "year": "",
                            "class_": "Bahn A - Lang",
                            "result": result_type.PersonRaceResult(),
                            "points": 0.091,
                        },
                    ],
                ),
            ],
        ],
    )
    print(data)
    assert data == [
        (
            class_a.name,
            [
                {
                    "first_name": "Angela",
                    "last_name": "Merkel",
                    "year": "",
                    "club": None,
                    "races": {0: Decimal("99.5")},
                    "organizer": {},
                    "sum": Decimal("99.5"),
                    "rank": 1,
                },
                {
                    "first_name": "Gerd",
                    "last_name": "Müller",
                    "year": "",
                    "club": None,
                    "races": {1: Decimal("9.1")},
                    "organizer": {},
                    "sum": Decimal("9.1"),
                    "rank": 2,
                },
            ],
        ),
    ]


def test_ranking_3():
    class_a = Class_(name="Bahn A - Lang")
    data = series.build_total_results(
        settings=series_type.Settings(),
        list_of_results=[
            [
                (
                    class_a,
                    [
                        {
                            "first_name": "Angela",
                            "last_name": "Merkel",
                            "gender": "",
                            "year": "",
                            "class_": "Bahn A - Lang",
                            "result": result_type.PersonRaceResult(),
                            "points": 0.65,
                        },
                    ],
                )
            ],
            [
                (
                    class_a,
                    [
                        {
                            "first_name": "Gerd",
                            "last_name": "Müller",
                            "gender": "",
                            "year": "",
                            "class_": "Bahn A - Lang",
                            "result": result_type.PersonRaceResult(),
                            "points": 0.95,
                        },
                        {
                            "first_name": "Angela",
                            "last_name": "Merkel",
                            "gender": "",
                            "year": "",
                            "class_": "Bahn A - Lang",
                            "result": result_type.PersonRaceResult(),
                            "points": 0.3,
                        },
                    ],
                ),
            ],
        ],
    )
    print(data)
    assert data == [
        (
            class_a.name,
            [
                {
                    "first_name": "Angela",
                    "last_name": "Merkel",
                    "year": "",
                    "club": None,
                    "races": {0: Decimal("65.0"), 1: Decimal("30.0")},
                    "organizer": {},
                    "sum": Decimal("95.0"),
                    "rank": 1,
                },
                {
                    "first_name": "Gerd",
                    "last_name": "Müller",
                    "year": "",
                    "club": None,
                    "races": {1: Decimal("95.0")},
                    "organizer": {},
                    "sum": Decimal("95.0"),
                    "rank": 1,
                },
            ],
        ),
    ]


def test_if_sum_is_0_then_rank_is_None():
    class_a = Class_(name="Bahn A - Lang")
    data = series.build_total_results(
        settings=series_type.Settings(),
        list_of_results=[
            [
                (
                    class_a,
                    [
                        {
                            "first_name": "Angela",
                            "last_name": "Merkel",
                            "gender": "",
                            "year": "",
                            "class_": "Bahn A - Lang",
                            "result": result_type.PersonRaceResult(),
                            "points": 0.65,
                        },
                    ],
                )
            ],
            [
                (
                    class_a,
                    [
                        {
                            "first_name": "Gerd",
                            "last_name": "Müller",
                            "gender": "",
                            "year": "",
                            "class_": "Bahn A - Lang",
                            "result": result_type.PersonRaceResult(),
                            "points": 0,
                        },
                        {
                            "first_name": "Angela",
                            "last_name": "Merkel",
                            "gender": "",
                            "year": "",
                            "class_": "Bahn A - Lang",
                            "result": result_type.PersonRaceResult(),
                            "points": 0.3,
                        },
                    ],
                ),
            ],
        ],
    )
    print(data)
    assert data == [
        (
            class_a.name,
            [
                {
                    "first_name": "Angela",
                    "last_name": "Merkel",
                    "year": "",
                    "club": None,
                    "races": {0: Decimal("65.0"), 1: Decimal("30.0")},
                    "organizer": {},
                    "sum": Decimal("95.0"),
                    "rank": 1,
                },
                {
                    "first_name": "Gerd",
                    "last_name": "Müller",
                    "year": "",
                    "club": None,
                    "races": {1: Decimal("0.0")},
                    "organizer": {},
                    "sum": Decimal("0.0"),
                    "rank": None,
                },
            ],
        ),
    ]



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import argparse
import configparser
import datetime
import pathlib
import json
import time
import threading
import bz2
import ssl
import queue
from typing import Dict
from typing import Optional

import websocket
import jsonschema
import sireader
import serial.tools.list_ports


#
# ConfigFile:
#
#  [Server]
#  ssl_cert = cert/cert.pem
#  ssl_key = cert/privkey.pem
#
#  [Cardreader]
#  host = localhost
#  ssl_cert = cert/cert.pem
#  ssl_verify = true
#  key = 4711
#  serial_number =
#


class WebSocketClient(threading.Thread):
    def __init__(
        self,
        host: str = "localhost",
        port: int = 8081,
        key: str = "",
        ssl_cert="cert/cert.pem",
        ssl_verify: bool = True,
    ):
        super().__init__()
        self.daemon = True
        self.host = host
        self.port = port
        self.key = key
        self.ssl_cert = ssl_cert
        self.ssl_verify = ssl_verify
        self.ws = None
        self.opened = False
        self.queue = queue.Queue()
        self.entry_type = "readerDisconnected"
        self.entry_time = datetime.datetime.now()
        self.card = None

    def run(self):
        headers = {
            "Content-Type": "application/octet-stream",
            "X-Event-Key": self.key,
            "X-Suffix": ".json",
        }
        sslopt = {"cert_reqs": ssl.CERT_REQUIRED}

        if not self.ssl_verify:
            # disable ssl verification: sslopt = {'cert_reqs': ssl.CERT_NONE}
            # disable hostname verification: sslopt = {'check_hostname': False}
            sslopt["cert_reqs"] = ssl.CERT_NONE
        if self.ssl_cert is not None:
            sslopt["ca_certs"] = self.ssl_cert

        uri = f"wss://{self.host}:{str(self.port)}/cardreader"
        while True:
            print(f"Trying to connect to {uri} ...")
            self.ws = websocket.WebSocketApp(
                url=uri,
                header=headers,
                on_open=self.on_open,
                on_close=self.on_close,
                on_message=self.on_message,
                on_error=self.on_error,
            )
            self.ws.run_forever(sslopt=sslopt)
            time.sleep(5)

    def set_state(
        self, entry_type: str, entry_time: datetime.datetime, card: Optional[str] = None
    ) -> Dict:
        self.entry_type = entry_type
        self.entry_time = entry_time
        self.card = card
        # send a cardreader state message
        item = {
            "entryType": self.entry_type,
            "entryTime": self.entry_time.astimezone().isoformat(),
        }
        if self.card is not None:
            item["controlCard"] = self.card
        return self.send_and_receive(item=item)

    def send_and_receive(self, item: Dict, timeout: Optional[int] = None) -> Dict:
        self.clear()
        data = bz2.compress(json.dumps(item).encode())
        self.send(data)
        return self.receive(timeout=timeout)

    def send(self, data: bytes) -> None:
        if self.opened:
            self.ws.send(data, opcode=2)
        else:
            print("Not connected to server - could not send data")

    def receive(self, timeout: Optional[int] = None) -> Dict:
        if self.opened:
            return json.loads(self.queue.get(timeout=timeout))
        else:
            return {}

    def clear(self) -> None:
        while not self.queue.empty():
            self.queue.get()

    def on_close(self, wsapp, close_status_code, close_msg):
        if self.opened:
            print("Connection closed")
        self.opened = False

    def on_open(self, wsapp):
        if not self.opened:
            print("Connection established")
            self.opened = True
            item = {
                "entryType": self.entry_type,
                "entryTime": self.entry_time.astimezone().isoformat(),
            }
            data = bz2.compress(json.dumps(item).encode())
            self.send(data=data)

    def on_message(self, wsapp, message):
        self.queue.put(message)

    def on_error(self, wsapp, message):
        print(message)


class Cardreader:
    data_path = (
        pathlib.Path(__file__).resolve().parent / "schema" / "cardreader_log.json"
    )
    with open(data_path, "r") as f:
        schema_cardreader_log = json.loads(f.read())

    def __init__(self, webSocketClient: WebSocketClient, serial_number: str = ""):
        self.webSocketClient = webSocketClient
        self.serial_number = serial_number

    def convert(self, card_type: str, card_data: Dict) -> Dict:
        item = {
            "entryType": "cardRead",
            "entryTime": datetime.datetime.now().astimezone().isoformat(),
            "cardType": card_type,
            "controlCard": str(card_data["card_number"]),
        }

        if card_data.get("start", None) is not None:
            item["startTime"] = card_data["start"].astimezone().isoformat()
        if card_data.get("finish", None) is not None:
            item["finishTime"] = card_data["finish"].astimezone().isoformat()
        if card_data.get("check", None) is not None:
            item["checkTime"] = card_data["check"].astimezone().isoformat()
        if card_data.get("clear", None) is not None:
            item["clearTime"] = card_data["clear"].astimezone().isoformat()

        item["punches"] = []
        for p in card_data["punches"]:
            item["punches"].append(
                {"controlCode": str(p[0]), "punchTime": p[1].astimezone().isoformat()}
            )

        return item

    def protocol(self, item: Dict) -> None:
        date_str = datetime.date.today().isoformat()
        with open(f"cardreader-{date_str}.log", "a") as f:
            f.write(json.dumps(item) + "\n")

    def connect(self) -> sireader.SIReaderReadout:
        errors = ""
        for port in serial.tools.list_ports.grep("sportident"):
            try:
                si = sireader.SIReaderReadout(port.device)
                if not self.serial_number or self.serial_number == port.serial_number:
                    print(f"SI Reader found, serial number: {port.serial_number}")
                    return si
                else:
                    print(f"SI Reader found, but serial number: {port.serial_number}")

            except (sireader.SIReaderException, sireader.SIReaderTimeout) as msg:
                errors += f"port: {port.device}: {msg}\n"
        else:
            errors = "No SI Reader found" if errors == "" else errors
        raise sireader.SIReaderException(
            f"No SI Reader found. Possible reasons: {errors}"
        )

    def reader(self) -> None:
        # connect to base station, the station is automatically detected,
        # if this does not work, give the path to the port as an argument
        # see the pyserial documentation for further information.
        si = self.connect()

        # check extended protocol mode
        extended_protocol = True
        try:
            si.poll_sicard()
        except sireader.SIReaderException:
            extended_protocol = False
            # change to extended protocol mode
            si.set_extended_protocol(True)

        try:
            r = self.webSocketClient.set_state(
                entry_type="readerConnected",
                entry_time=datetime.datetime.now(),
            )

            while True:
                # wait for a card to be inserted into the reader
                while not si.poll_sicard() or si.sicard is None:
                    time.sleep(0.2)

                try:
                    # some properties are now set
                    card_number = si.sicard  #  8320666
                    card_type = si.cardtype  #  'SI10'
                    self.webSocketClient.set_state(
                        entry_type="cardInserted",
                        entry_time=datetime.datetime.now(),
                        card=str(si.sicard),
                    )

                    # read out card data
                    #
                    # {'card_number': 219412,
                    #  'start': datetime.datetime(2021, 5, 18, 16, 31, 19),
                    #  'finish': datetime.datetime(2021, 5, 18, 16, 31, 50),
                    #  'check': datetime.datetime(2021, 5, 18, 16, 31, 18),
                    #  'clear': None,
                    #  'punches': [(141, datetime.datetime(2021, 5, 18, 16, 31, 25)),
                    #              (143, datetime.datetime(2021, 5, 18, 16, 31, 31)),
                    #              (145, datetime.datetime(2021, 5, 18, 16, 31, 38)),
                    #              (143, datetime.datetime(2021, 5, 18, 16, 31, 44)),
                    #             ],
                    # }
                    #
                    card_data = si.read_sicard()

                    item = self.convert(card_type=card_type, card_data=card_data)
                    self.protocol(item=item)
                    print("")
                    print("Entry time:  ", item["entryTime"])
                    print("Card number: ", item["controlCard"])
                    print("Start time:  ", item.get("startTime", None))
                    print("Finish time: ", item.get("finishTime", None))
                    print("Controls:    ", len(item.get("punches", 0)))

                    r = self.webSocketClient.send_and_receive(item=item)

                    if "status" in r:
                        print(r["status"])
                        # beep
                        si.ack_sicard()
                except sireader.SIReaderCardChanged:
                    self.webSocketClient.set_state(
                        entry_type="cardRemoved",
                        entry_time=datetime.datetime.now(),
                    )

        finally:
            try:
                if not extended_protocol:
                    # change back to basic protocl
                    si.set_extended_protocol(False)
            finally:
                si.disconnect()
                # send readerDisconnected message
                # send readerDisconnected message
                r = self.webSocketClient.set_state(
                    entry_type="readerDisconnected",
                    entry_time=datetime.datetime.now(),
                )

    def process_cards(self) -> None:
        print("#### process cards ###")
        while True:
            try:
                self.reader()
            except (sireader.SIReaderException, sireader.SIReaderTimeout, OSError) as e:
                print(f"SIReader or OSError exception:\n{str(e)}")
                time.sleep(5)

    def process_log(self, cardreader_log: pathlib.Path) -> None:
        send_all_entries = False
        with open(cardreader_log, "r") as f:
            r = self.webSocketClient.set_state(
                entry_type="readerConnected",
                entry_time=datetime.datetime.now(),
            )

            line = f.readline()
            while line != "":
                if not send_all_entries:
                    inp = input("Weiter? ")
                else:
                    inp = ""

                if inp == "q":
                    break
                elif inp == "a":
                    send_all_entries = True
                elif inp == "c":
                    self.webSocketClient.set_state(
                        entry_type="readerConnected",
                        entry_time=datetime.datetime.now(),
                    )
                elif inp == "d":
                    self.webSocketClient.set_state(
                        entry_type="readerDisconnected",
                        entry_time=datetime.datetime.now(),
                    )
                elif inp == "i":
                    self.webSocketClient.set_state(
                        entry_type="cardInserted",
                        entry_time=datetime.datetime.now(),
                        card=json.loads(line)["controlCard"],
                    )
                elif inp == "r":
                    self.webSocketClient.set_state(
                        entry_type="cardRemoved",
                        entry_time=datetime.datetime.now(),
                    )
                else:
                    item = json.loads(line)
                    line = f.readline()

                    jsonschema.validate(item, self.schema_cardreader_log)
                    print(f"{str(item.get('controlCard', item['entryType']))} ...")

                    r = self.webSocketClient.send_and_receive(item=item)
                    print(r)


def main() -> Optional[int]:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-c",
        "--config",
        type=pathlib.Path,
        default=pathlib.Path.home() / ".ooresults" / "config.ini",
    )
    parser.add_argument("-f", "--file", type=pathlib.Path)
    parser.add_argument("-v", "--verbose", action="store_true")
    args = parser.parse_args()

    config_file = pathlib.Path(args.config)
    if not config_file.exists():
        parser.error(f"File {str(config_file)} not found")

    config = configparser.ConfigParser()
    config.read(config_file)

    # check config file
    #
    #  [Cardreader]
    #  host = localhost
    #  ssl_cert = cert/cert.pem
    #  ssl_verify = true
    #  key = 4711
    #  serial_number =
    #
    try:
        if not config.has_section("Cardreader"):
            raise RuntimeError("Section [Cardreader] missing")
        if not config.has_option("Cardreader", "host"):
            raise RuntimeError("Section [Cardreader]: option 'host' missing")
        if not config.has_option("Cardreader", "key"):
            raise RuntimeError("Section [Cardreader]: option 'key' missing")
    except RuntimeError as e:
        print(f"Error in file {str(config_file)}:")
        print(f"  {str(e)}")
        return 2

    try:
        host = config["Cardreader"]["host"]
        key = config["Cardreader"]["key"]
        ssl_cert = config.get("Cardreader", "ssl_cert", fallback=None)
        serial_number = config.get("Cardreader", "serial_number", fallback=None)
        try:
            ssl_verify = config.getboolean("Cardreader", "ssl_verify", fallback=True)
        except ValueError:
            parser.error(
                "Allowed values for 'ssl_verify' are 'true', 'false', 'on', 'off', 'yes', 'no'"
            )

        # check ssl options
        if ssl_cert and not pathlib.Path(ssl_cert).exists():
            parser.error(f'Certificate file "{ssl_cert}" not found')

        print(f"Host:   {str(host)}")
        print(f"Cert:   {str(ssl_cert)}")
        print(f"Verify: {str(ssl_verify)}")
        print(f"Key:    {str(key)}")
        print(f"Serial: {str(serial_number)}")
        print("")

        # websocket.enableTrace(True)
        webSocketClient = WebSocketClient(
            host=host, key=key, ssl_cert=ssl_cert, ssl_verify=ssl_verify
        )
        webSocketClient.start()

        cardreader = Cardreader(
            webSocketClient=webSocketClient, serial_number=serial_number
        )
        if args.file is None:
            cardreader.process_cards()
        else:
            cardreader.process_log(cardreader_log=args.file)

    except KeyboardInterrupt:
        pass



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import datetime
import pathlib
import logging
from typing import Optional

import tzlocal
import web
from web.utils import Storage

from ooresults.handler import model
from ooresults.repo.repo import ClassUsedError
from ooresults.repo.repo import EventNotFoundError
from ooresults.repo.repo import ConstraintError
from ooresults.repo.class_params import ClassParams
from ooresults.plugins import iof_class_list


templates = pathlib.Path(__file__).resolve().parent.parent / "templates"
render = web.template.render(templates, globals={"str": str})


def update(event_id: int):
    classes_list = model.get_classes(event_id=event_id)
    event = list(model.get_event(event_id))
    if event == []:
        return render.classes_table({}, classes_list)
    else:
        return render.classes_table(event[0], classes_list)


class Update:
    def POST(self):
        """Update data"""
        data = web.input()
        event_id = int(data.event_id) if data.event_id != "" else -1
        return update(event_id)


class Import:
    def POST(self):
        """Import classes"""
        data = web.input()
        event_id = int(data.event_id) if data.event_id != "" else -1
        try:
            if data.cls_import == "cls.import.1":
                classes = iof_class_list.parse_class_list(data.browse1)
                model.import_classes(event_id=event_id, classes=classes)

        except EventNotFoundError:
            raise web.conflict("No event selected or event deleted")
        except Exception as e:
            raise web.Conflict(str(e))

        return update(event_id)


class Export:
    def POST(self):
        """Export classes"""
        data = web.input()
        event_id = int(data.event_id) if data.event_id != "" else -1
        try:
            if data.cls_export == "cls.export.1":
                classes = model.get_classes(event_id=event_id)
                content = iof_class_list.create_class_list(classes)

        except KeyError:
            raise web.conflict("Entry deleted")
        except:
            logging.exception("Internal server error")
            raise

        return content


class Add:
    def parse_start_time(
        self, item: str, event_date: datetime.date
    ) -> Optional[datetime.datetime]:
        if item != "":
            format = "%H:%M:%S" if item.count(":") == 2 else "%M:%S"
            tz = tzlocal.get_localzone()
            dt = datetime.datetime.combine(
                date=event_date,
                time=datetime.datetime.strptime(item, format).time(),
                tzinfo=tz,
            )
            print(">>> ", dt)
            return dt
        else:
            return None

    def POST(self):
        """Add or edit entry"""
        data = web.input()
        print(data)
        event_id = int(data.event_id) if data.event_id != "" else -1
        event = model.get_event(id=event_id)[0]

        params = ClassParams()
        params.otype = data.get("type", "standard")
        params.using_start_control = data.get("startControl", "if_punched")
        params.mass_start = self.parse_start_time(data.massStart, event.date)
        params.apply_handicap_rule = data.get("handicap", "0") == "1"
        if data.get("timeLimit", "") != "":
            m, _, s = data["timeLimit"].partition(":")
            params.time_limit = 60 * int(m) + int(s)
        else:
            params.time_limit = None
        params.penalty_controls = (
            int(data.get("penaltyControls", ""))
            if data.get("penaltyControls", "") != ""
            else None
        )
        params.penalty_overtime = (
            int(data.get("penaltyOvertime", ""))
            if data.get("penaltyOvertime", "") != ""
            else None
        )

        voided_legs = (
            data.get("voided_legs", "").split(",")
            if data.get("voided_legs", "") != ""
            else []
        )
        for v in voided_legs:
            # there should be two controls separated by '-'
            controls = v.split("-")
            if len(controls) == 2:
                if (controls[0].strip(), controls[1].strip()) not in params.voided_legs:
                    params.voided_legs.append(
                        (controls[0].strip(), controls[1].strip())
                    )
            else:
                raise web.conflict(
                    'Format of voided legs incorrect, use "c1-c2, c3-c4, ..."'
                )

        try:
            course_id = int(data.course) if data.course != "" else None
            if data.id == "":
                model.add_class(
                    event_id=event_id,
                    name=data.name,
                    short_name=data.short_name if data.short_name != "" else None,
                    course_id=course_id,
                    params=params,
                )

            else:
                model.update_class(
                    id=int(data.id),
                    event_id=event_id,
                    name=data.name,
                    short_name=data.short_name if data.short_name != "" else None,
                    course_id=int(data.course) if data.course != "" else None,
                    params=params,
                )

        except EventNotFoundError:
            raise web.conflict("No event selected or event deleted")
        except ConstraintError as e:
            raise web.conflict(str(e))
        except KeyError:
            raise web.conflict("Class deleted")
        except:
            logging.exception("Internal server error")
            raise

        return update(event_id)


class Delete:
    def POST(self):
        """Delete entry"""
        data = web.input()
        print(data)
        event_id = int(data.event_id) if data.event_id != "" else -1
        try:
            model.delete_class(id=int(data.id))
            return update(event_id)

        except ClassUsedError:
            raise web.conflict("Class used in entries")
        except:
            logging.exception("Internal server error")
            raise


class FillEditForm:
    def POST(self):
        """Query data to fill add or edit form"""
        data = web.input()
        event_id = int(data.event_id) if data.event_id != "" else -1
        try:
            if data.id == "":
                class_ = Storage(
                    {
                        "id": "",
                        "name": "",
                        "short_name": None,
                        "course_id": None,
                        "params": ClassParams(),
                    }
                )
            else:
                class_ = model.get_class(id=int(data.id))[0]

        except EventNotFoundError:
            raise web.conflict("No event selected or event deleted")
        except KeyError:
            raise web.conflict("Class deleted")
        except:
            logging.exception("Internal server error")
            raise

        courses = model.get_courses(event_id=event_id)
        return render.add_class(class_, courses)



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from ooresults.handler import results
from ooresults.repo import result_type
from ooresults.repo.result_type import ResultStatus
from ooresults.repo.class_params import ClassParams


class Class_:
    def __init__(self, name: str, params: ClassParams):
        self.name = name
        self.params = params


def test_ranking_with_one_class():
    class_a = Class_(name="Bahn A - Lang", params=ClassParams())
    data = results.build_results(
        classes=[
            class_a,
        ],
        results=[
            {
                "first_name": "Angela",
                "last_name": "Merkel",
                "gender": "",
                "year": "",
                "not_competing": False,
                "class_": "Bahn A - Lang",
                "result": result_type.PersonRaceResult(
                    status=ResultStatus.OK, time=9876
                ),
            },
            {
                "first_name": "Claudia",
                "last_name": "Merkel",
                "gender": "",
                "year": "",
                "not_competing": False,
                "class_": "Bahn A - Lang",
                "result": result_type.PersonRaceResult(
                    status=ResultStatus.OK, time=2001
                ),
            },
            {
                "first_name": "Birgit",
                "last_name": "Merkel",
                "gender": "",
                "year": "",
                "not_competing": False,
                "class_": "Bahn A - Lang",
                "result": result_type.PersonRaceResult(
                    status=ResultStatus.MISSING_PUNCH
                ),
            },
            {
                "first_name": "Birgit",
                "last_name": "Derkel",
                "gender": "",
                "year": "",
                "not_competing": False,
                "class_": "Bahn A - Lang",
                "result": result_type.PersonRaceResult(
                    status=ResultStatus.DID_NOT_START
                ),
            },
        ],
    )
    assert data == [
        (
            class_a,
            [
                {
                    "first_name": "Claudia",
                    "last_name": "Merkel",
                    "class_": "Bahn A - Lang",
                    "gender": "",
                    "year": "",
                    "not_competing": False,
                    "result": result_type.PersonRaceResult(
                        status=ResultStatus.OK, time=2001
                    ),
                    "rank": 1,
                    "points": 1.0,
                },
                {
                    "first_name": "Angela",
                    "last_name": "Merkel",
                    "class_": "Bahn A - Lang",
                    "gender": "",
                    "year": "",
                    "not_competing": False,
                    "result": result_type.PersonRaceResult(
                        status=ResultStatus.OK, time=9876
                    ),
                    "rank": 2,
                    "points": 2001 / 9876,
                },
                {
                    "first_name": "Birgit",
                    "last_name": "Merkel",
                    "class_": "Bahn A - Lang",
                    "gender": "",
                    "year": "",
                    "not_competing": False,
                    "result": result_type.PersonRaceResult(
                        status=ResultStatus.MISSING_PUNCH, time=None
                    ),
                    "rank": None,
                    "points": 0,
                },
                {
                    "first_name": "Birgit",
                    "last_name": "Derkel",
                    "class_": "Bahn A - Lang",
                    "gender": "",
                    "year": "",
                    "not_competing": False,
                    "result": result_type.PersonRaceResult(
                        status=ResultStatus.DID_NOT_START, time=None
                    ),
                    "rank": None,
                },
            ],
        )
    ]


def test_ranking_with_two_winners():
    class_a = Class_(name="Bahn A - Lang", params=ClassParams())
    data = results.build_results(
        classes=[
            class_a,
        ],
        results=[
            {
                "first_name": "Angela",
                "last_name": "Merkel",
                "gender": "",
                "year": "",
                "not_competing": False,
                "class_": "Bahn A - Lang",
                "result": result_type.PersonRaceResult(
                    status=ResultStatus.MISSING_PUNCH, time=9876
                ),
            },
            {
                "first_name": "Claudia",
                "last_name": "Merkel",
                "gender": "",
                "year": "",
                "not_competing": False,
                "class_": "Bahn A - Lang",
                "result": result_type.PersonRaceResult(
                    status=ResultStatus.OK, time=2001
                ),
            },
            {
                "first_name": "Birgit",
                "last_name": "Merkel",
                "gender": "",
                "year": "",
                "not_competing": False,
                "class_": "Bahn A - Lang",
                "result": result_type.PersonRaceResult(
                    status=ResultStatus.OK, time=2001
                ),
            },
        ],
    )
    assert data == [
        (
            class_a,
            [
                {
                    "first_name": "Birgit",
                    "last_name": "Merkel",
                    "class_": "Bahn A - Lang",
                    "gender": "",
                    "year": "",
                    "not_competing": False,
                    "result": result_type.PersonRaceResult(
                        status=ResultStatus.OK, time=2001
                    ),
                    "rank": 1,
                    "points": 1.0,
                },
                {
                    "first_name": "Claudia",
                    "last_name": "Merkel",
                    "class_": "Bahn A - Lang",
                    "gender": "",
                    "year": "",
                    "not_competing": False,
                    "result": result_type.PersonRaceResult(
                        status=ResultStatus.OK, time=2001
                    ),
                    "rank": 1,
                    "points": 1.0,
                },
                {
                    "first_name": "Angela",
                    "last_name": "Merkel",
                    "class_": "Bahn A - Lang",
                    "gender": "",
                    "year": "",
                    "not_competing": False,
                    "result": result_type.PersonRaceResult(
                        status=ResultStatus.MISSING_PUNCH, time=9876
                    ),
                    "rank": None,
                    "points": 0,
                },
            ],
        )
    ]


def test_ranking_entries_with_same_time_are_orderd_by_name():
    class_a = Class_(name="Bahn A - Lang", params=ClassParams())
    data = results.build_results(
        classes=[
            class_a,
        ],
        results=[
            {
                "first_name": "Angela",
                "last_name": "Merkel",
                "gender": "",
                "year": "",
                "not_competing": False,
                "class_": "Bahn A - Lang",
                "result": result_type.PersonRaceResult(
                    status=ResultStatus.OK, time=2001
                ),
            },
            {
                "first_name": "Claudia",
                "last_name": "Merkel",
                "gender": "",
                "year": "",
                "not_competing": False,
                "class_": "Bahn A - Lang",
                "result": result_type.PersonRaceResult(
                    status=ResultStatus.OK, time=2001
                ),
            },
            {
                "first_name": "Birgit",
                "last_name": "Derkel",
                "gender": "",
                "year": "",
                "not_competing": False,
                "class_": "Bahn A - Lang",
                "result": result_type.PersonRaceResult(
                    status=ResultStatus.OK, time=2001
                ),
            },
            {
                "first_name": "Birgit",
                "last_name": "Merkel",
                "gender": "",
                "year": "",
                "not_competing": False,
                "class_": "Bahn A - Lang",
                "result": result_type.PersonRaceResult(
                    status=ResultStatus.OK, time=2001
                ),
            },
        ],
    )
    print(data)
    assert data == [
        (
            class_a,
            [
                {
                    "first_name": "Birgit",
                    "last_name": "Derkel",
                    "class_": "Bahn A - Lang",
                    "gender": "",
                    "year": "",
                    "not_competing": False,
                    "result": result_type.PersonRaceResult(
                        status=ResultStatus.OK, time=2001
                    ),
                    "rank": 1,
                    "points": 1.0,
                },
                {
                    "first_name": "Angela",
                    "last_name": "Merkel",
                    "class_": "Bahn A - Lang",
                    "gender": "",
                    "year": "",
                    "not_competing": False,
                    "result": result_type.PersonRaceResult(
                        status=ResultStatus.OK, time=2001
                    ),
                    "rank": 1,
                    "points": 1.0,
                },
                {
                    "first_name": "Birgit",
                    "last_name": "Merkel",
                    "class_": "Bahn A - Lang",
                    "gender": "",
                    "year": "",
                    "not_competing": False,
                    "result": result_type.PersonRaceResult(
                        status=ResultStatus.OK, time=2001
                    ),
                    "rank": 1,
                    "points": 1.0,
                },
                {
                    "first_name": "Claudia",
                    "last_name": "Merkel",
                    "class_": "Bahn A - Lang",
                    "gender": "",
                    "year": "",
                    "not_competing": False,
                    "result": result_type.PersonRaceResult(
                        status=ResultStatus.OK, time=2001
                    ),
                    "rank": 1,
                    "points": 1.0,
                },
            ],
        )
    ]


def test_ranking_entries_with_same_time_are_orderd_by_name_2():
    class_a = Class_(name="Bahn A - Lang", params=ClassParams(otype="score"))
    data = results.build_results(
        classes=[
            class_a,
        ],
        results=[
            {
                "first_name": "Claudia",
                "last_name": "Merkel",
                "gender": "",
                "year": "",
                "not_competing": False,
                "class_": "Bahn A - Lang",
                "result": result_type.PersonRaceResult(
                    status=ResultStatus.OK, time=2000, extensions={"score": 15.6}
                ),
            },
            {
                "first_name": "Tanja",
                "last_name": "Terkel",
                "gender": "",
                "year": "",
                "not_competing": False,
                "class_": "Bahn A - Lang",
                "result": result_type.PersonRaceResult(
                    status=ResultStatus.OK, time=2001, extensions={"score": 15.6}
                ),
            },
            {
                "first_name": "Angela",
                "last_name": "Merkel",
                "gender": "",
                "year": "",
                "not_competing": False,
                "class_": "Bahn A - Lang",
                "result": result_type.PersonRaceResult(
                    status=ResultStatus.OK, time=2000, extensions={"score": 15.6}
                ),
            },
            {
                "first_name": "Birgit",
                "last_name": "Derkel",
                "gender": "",
                "year": "",
                "not_competing": False,
                "class_": "Bahn A - Lang",
                "result": result_type.PersonRaceResult(
                    status=ResultStatus.OK, time=2001, extensions={"score": 15.6}
                ),
            },
        ],
    )
    assert data == [
        (
            class_a,
            [
                {
                    "first_name": "Angela",
                    "last_name": "Merkel",
                    "class_": "Bahn A - Lang",
                    "gender": "",
                    "year": "",
                    "not_competing": False,
                    "result": result_type.PersonRaceResult(
                        status=ResultStatus.OK, time=2000, extensions={"score": 15.6}
                    ),
                    "rank": 1,
                },
                {
                    "first_name": "Claudia",
                    "last_name": "Merkel",
                    "class_": "Bahn A - Lang",
                    "gender": "",
                    "year": "",
                    "not_competing": False,
                    "result": result_type.PersonRaceResult(
                        status=ResultStatus.OK, time=2000, extensions={"score": 15.6}
                    ),
                    "rank": 1,
                },
                {
                    "first_name": "Birgit",
                    "last_name": "Derkel",
                    "class_": "Bahn A - Lang",
                    "gender": "",
                    "year": "",
                    "not_competing": False,
                    "result": result_type.PersonRaceResult(
                        status=ResultStatus.OK, time=2001, extensions={"score": 15.6}
                    ),
                    "rank": 3,
                },
                {
                    "first_name": "Tanja",
                    "last_name": "Terkel",
                    "class_": "Bahn A - Lang",
                    "gender": "",
                    "year": "",
                    "not_competing": False,
                    "result": result_type.PersonRaceResult(
                        status=ResultStatus.OK, time=2001, extensions={"score": 15.6}
                    ),
                    "rank": 3,
                },
            ],
        )
    ]


def test_ranking_with_not_competing_runners():
    class_a = Class_(name="Bahn A - Lang", params=ClassParams())
    data = results.build_results(
        classes=[
            class_a,
        ],
        results=[
            {
                "first_name": "Angela",
                "last_name": "Merkel",
                "gender": "",
                "year": "",
                "not_competing": False,
                "class_": "Bahn A - Lang",
                "result": result_type.PersonRaceResult(
                    status=ResultStatus.OK, time=9876
                ),
            },
            {
                "first_name": "Claudia",
                "last_name": "Merkel",
                "gender": "",
                "year": "",
                "not_competing": True,
                "class_": "Bahn A - Lang",
                "result": result_type.PersonRaceResult(
                    status=ResultStatus.OK, time=2001
                ),
            },
            {
                "first_name": "Birgit",
                "last_name": "Merkel",
                "gender": "",
                "year": "",
                "not_competing": True,
                "class_": "Bahn A - Lang",
                "result": result_type.PersonRaceResult(
                    status=ResultStatus.MISSING_PUNCH
                ),
            },
            {
                "first_name": "Birgit",
                "last_name": "Derkel",
                "gender": "",
                "year": "",
                "not_competing": False,
                "class_": "Bahn A - Lang",
                "result": result_type.PersonRaceResult(
                    status=ResultStatus.DID_NOT_START
                ),
            },
        ],
    )
    print(data[0])
    assert data == [
        (
            class_a,
            [
                {
                    "first_name": "Angela",
                    "last_name": "Merkel",
                    "class_": "Bahn A - Lang",
                    "gender": "",
                    "year": "",
                    "not_competing": False,
                    "result": result_type.PersonRaceResult(
                        status=ResultStatus.OK, time=9876
                    ),
                    "rank": 1,
                    "points": 2001 / 2001,
                },
                {
                    "first_name": "Claudia",
                    "last_name": "Merkel",
                    "class_": "Bahn A - Lang",
                    "gender": "",
                    "year": "",
                    "not_competing": True,
                    "result": result_type.PersonRaceResult(
                        status=ResultStatus.OK, time=2001
                    ),
                    "rank": None,
                    "points": 0,
                },
                {
                    "first_name": "Birgit",
                    "last_name": "Merkel",
                    "class_": "Bahn A - Lang",
                    "gender": "",
                    "year": "",
                    "not_competing": True,
                    "result": result_type.PersonRaceResult(
                        status=ResultStatus.MISSING_PUNCH, time=None
                    ),
                    "rank": None,
                    "points": 0,
                },
                {
                    "first_name": "Birgit",
                    "last_name": "Derkel",
                    "class_": "Bahn A - Lang",
                    "gender": "",
                    "year": "",
                    "not_competing": False,
                    "result": result_type.PersonRaceResult(
                        status=ResultStatus.DID_NOT_START, time=None
                    ),
                    "rank": None,
                },
            ],
        )
    ]



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import dataclasses
from datetime import datetime
from typing import List
from typing import Optional
from typing import Tuple


@dataclasses.dataclass
class ClassParams:
    otype: str = "standard"
    using_start_control: str = "if_punched"
    mass_start: Optional[datetime] = None
    time_limit: Optional[int] = None
    penalty_controls: Optional[int] = None
    penalty_overtime: Optional[int] = None
    apply_handicap_rule: bool = False
    voided_legs: List[Tuple[str, str]] = dataclasses.field(default_factory=list)

    #
    # possible values for using_start_control
    #   if_punched: if the start control station is punched, the punching time is used,
    #               if the start control station is not punched, the start time of the competitor is used
    #   yes: if the start control station is not punched, the race status is 'MissingPunch'
    #   no:  a punched start control station is ignored
    #
    # possible values for otype:
    #   standard: classic orienteering (control stations must be punched in given order)
    #   net:      control stations can be punched in arbitrary order
    #   score:
    #

    def __contains__(self, key):
        return hasattr(self, key)

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        return setattr(self, key, value)

    def get(self, key, default):
        return getattr(self, key, default)



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import dataclasses
from datetime import datetime
import enum
from typing import List
from typing import Optional
from typing import Dict

from ooresults.handler import handicap
from ooresults.repo.class_params import ClassParams


@dataclasses.dataclass
class SplitTime:
    control_code: str
    punch_time: Optional[datetime] = None
    time: Optional[int] = None
    status: Optional[str] = None
    leg_voided: bool = False

    def recalculate_time(self, start_time: Optional[datetime]) -> None:
        """
        set time as difference between start_time and punch_time
        """
        if start_time is not None and self.punch_time is not None:
            diff = self.punch_time.replace(microsecond=0) - start_time.replace(
                microsecond=0
            )
            self.time = int(diff.total_seconds())

    def __contains__(self, key):
        return hasattr(self, key)

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        return setattr(self, key, value)

    def get(self, key, default):
        return getattr(self, key, default)


class ResultStatus(enum.Enum):
    INACTIVE = 0
    FINISHED = 2
    OK = 3
    MISSING_PUNCH = 4
    DID_NOT_START = 5
    DID_NOT_FINISH = 6
    OVER_TIME = 7
    DISQUALIFIED = 8


@dataclasses.dataclass
class PersonRaceResult:
    status: ResultStatus = ResultStatus.INACTIVE
    start_time: Optional[datetime] = None
    finish_time: Optional[datetime] = None
    punched_clear_time: Optional[datetime] = None
    punched_check_time: Optional[datetime] = None
    punched_start_time: Optional[datetime] = None
    punched_finish_time: Optional[datetime] = None
    time: Optional[int] = None
    split_times: List[SplitTime] = dataclasses.field(default_factory=list)
    last_leg_voided: bool = False
    extensions: Dict = dataclasses.field(default_factory=dict)

    #
    # keys of extensions are:
    #   factor: float            -- handicap factor
    #   penalties_controls: int  -- penalty for missing controls in seconds
    #   penalties_overtime: int  -- penalty for exceeding the time limit in seconds
    #   running_time: int        -- running time (= finish time - start time)
    #   score_controls: float    -- score points for controls
    #   score_overtime: float    -- score points to subtract for exceeding the time limit
    #   score: float             -- total score points (score_controls - score_overtime)
    #

    def __contains__(self, key):
        return hasattr(self, key)

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        return setattr(self, key, value)

    def get(self, key, default):
        return getattr(self, key, default)

    def has_punches(self) -> bool:
        return (
            self.punched_start_time is not None
            or self.punched_finish_time is not None
            or [p for p in self.split_times if p.status != "Missing"] != []
        )

    def voided_legs(self) -> List[str]:
        voided_legs = []
        c1 = "S"
        for split_time in [s for s in self.split_times if s.status != "Additional"]:
            c2 = split_time.control_code
            if split_time.leg_voided and f"{c1}-{c2}" not in voided_legs:
                voided_legs.append(f"{c1}-{c2}")
            c1 = c2
        c2 = "F"
        if self.last_leg_voided and f"{c1}-{c2}" not in voided_legs:
            voided_legs.append(f"{c1}-{c2}")
        return voided_legs

    def compute_result(
        self,
        controls: List[str],
        class_params: ClassParams,
        start_time: Optional[datetime] = None,
        year: Optional[int] = None,
        gender: Optional[str] = None,
    ) -> None:
        #
        # If list of controls is empty and result is not inactive or finished, no compution is done.
        # In all other cases a new result is computed, but:
        #
        # - if old status is disqualified the new status is disqualified
        # - if old status is inactive and no controls are punched the new status is inactive
        # - if old status is didnotstart and no controls are punched the new status is didnotstart
        # - if list of controls is empty and type is standard or net the new status is finished
        # - if list of controls is empty and type is score the new status is ok
        #

        # if list of controls is empty and result is not inactive or finished,
        # no compution is done
        if controls == [] and self.status not in [
            ResultStatus.INACTIVE,
            ResultStatus.FINISHED,
        ]:
            return

        # remove missing controls and reset results
        old_status = self.status
        split_times = [s for s in self.split_times if s.status != "Missing"]
        self.time = None
        self.status = ResultStatus.INACTIVE
        self.extensions = {}
        self.last_leg_voided = False
        for s in self.split_times:
            s.leg_voided = False

        # compute handicap factor
        handicap_factor = None
        if class_params.apply_handicap_rule:
            # use year of finish time to compute the age of the competitor
            h = handicap.Handicap()
            if self.punched_finish_time is not None and year is not None:
                handicap_factor = h.factor(
                    female=gender == "F", year=self.punched_finish_time.year - year
                )
            else:
                handicap_factor = h.factor(female=gender == "F", year=None)

        # compute start time
        if class_params.using_start_control == "yes":
            self.start_time = self.punched_start_time
        else:
            if (
                class_params.using_start_control == "if_punched"
                and self.punched_start_time is not None
            ):
                self.start_time = self.punched_start_time
            elif start_time is not None:
                self.start_time = start_time
            else:
                self.start_time = class_params.mass_start

        mp = self.start_time is None

        # compute finish time
        self.finish_time = self.punched_finish_time

        penalties_controls = 0
        if class_params.otype == "net":
            # controls can be visited in arbitrary order
            control_codes = controls.copy()
            self.split_times = split_times
            for p in self.split_times:
                if p.control_code in control_codes:
                    p.status = "OK"
                    p.recalculate_time(start_time=self.start_time)
                    control_codes = [c for c in control_codes if c != p.control_code]
                else:
                    p.status = "Additional"
                    p.recalculate_time(start_time=self.start_time)

            for control_code in control_codes:
                self.split_times.append(
                    SplitTime(control_code=control_code, status="Missing")
                )
                if class_params.penalty_controls is not None:
                    penalties_controls += class_params.penalty_controls
                else:
                    mp = True

        elif class_params.otype == "score":
            score_controls = 0
            # controls can be visited in arbitrary order
            control_codes = controls.copy()
            self.split_times = split_times
            for p in self.split_times:
                if p.control_code in control_codes:
                    p.status = "OK"
                    score_controls += 1
                    p.recalculate_time(start_time=self.start_time)
                    control_codes = [c for c in control_codes if c != p.control_code]
                else:
                    p.status = "Additional"
                    p.recalculate_time(start_time=self.start_time)

            for control_code in control_codes:
                self.split_times.append(
                    SplitTime(control_code=control_code, status="Missing")
                )

        else:
            # controls must be visited in the correct order
            self.split_times = []
            for control in controls:
                for i, p in enumerate(split_times):
                    if control == p.control_code:
                        for k in range(i):
                            p0 = split_times.pop(0)
                            p0.status = "Additional"
                            p0.recalculate_time(start_time=self.start_time)
                            self.split_times.append(p0)
                        p0 = split_times.pop(0)
                        p0.status = "OK"
                        p0.recalculate_time(start_time=self.start_time)
                        self.split_times.append(p0)
                        break
                else:
                    self.split_times.append(
                        SplitTime(control_code=control, status="Missing")
                    )
                    if class_params.penalty_controls is not None:
                        penalties_controls += class_params.penalty_controls
                    else:
                        mp = True

            for p in split_times:
                p.status = "Additional"
                p.recalculate_time(start_time=self.start_time)
                self.split_times.append(p)

        # compute result status
        # - if old status is disqualified the new status is disqualified
        # - if old status is inactive and no controls are punched the new status is inactive
        # - if old status is didnotstart and no controls are punched the new status is didnotstart
        # - if list of controls is empty and type is standard or net the new status is finished
        # - if list of controls is empty and type is score the new status is ok
        if old_status == ResultStatus.DISQUALIFIED:
            self.status = ResultStatus.DISQUALIFIED
        elif old_status == ResultStatus.INACTIVE and not self.has_punches():
            self.status = ResultStatus.INACTIVE
        elif old_status == ResultStatus.DID_NOT_START and not self.has_punches():
            self.status = ResultStatus.DID_NOT_START
        elif controls == [] and class_params.otype != "score":
            self.status = ResultStatus.FINISHED
        elif self.finish_time is None:
            self.status = ResultStatus.DID_NOT_FINISH
        else:
            # result status is DidNotFinish if the last three stations are missing
            missing = 0
            for p in reversed(self.split_times):
                if p.status == "OK":
                    break
                elif p.status == "Missing":
                    missing += 1
            if (
                class_params.otype == "standard"
                and class_params.penalty_overtime is None
                and missing >= 3
            ):
                self.status = ResultStatus.DID_NOT_FINISH
            else:
                self.status = ResultStatus.MISSING_PUNCH if mp else ResultStatus.OK

        # compute running time
        if self.start_time is not None and self.finish_time is not None:
            self.time = int((self.finish_time - self.start_time).total_seconds())

        # handle voided legs
        if class_params.otype == "standard" and class_params.voided_legs:
            # mark voided legs
            c1 = "S"
            for split_time in [s for s in self.split_times if s.status != "Additional"]:
                c2 = split_time.control_code
                if (c1, c2) in class_params.voided_legs:
                    split_time.leg_voided = True
                c1 = c2
            c2 = "F"
            if (c1, c2) in class_params.voided_legs:
                self.last_leg_voided = True

            # subtract time of voided legs from running time
            if self.time is not None:
                t1 = 0
                for split_time in [
                    s for s in self.split_times if s.status != "Additional"
                ]:
                    t2 = split_time.time
                    if split_time.leg_voided and t2 is not None and t1 is not None:
                        self.time -= t2 - t1
                    t1 = t2
                if self.last_leg_voided and t2 is not None:
                    self.time -= (
                        int((self.finish_time - self.start_time).total_seconds()) - t1
                    )

        if class_params.otype == "score":
            self.extensions["score_controls"] = score_controls

            # compute score for overtime
            if self.time is not None:
                score_overtime = 0
                overtime = self.time
                while overtime > class_params.time_limit:
                    overtime -= 60
                    score_overtime -= 1
                self.extensions["score_overtime"] = score_overtime

                # compute total score
                if class_params.apply_handicap_rule:
                    self.extensions["factor"] = handicap_factor
                    score = score_controls / handicap_factor + score_overtime
                else:
                    score = score_controls + score_overtime
                self.extensions["score"] = score
            else:
                self.extensions["score_overtime"] = None
                self.extensions["score"] = None

        else:
            # modify status if running time is to high
            if (
                class_params.time_limit is not None
                and class_params.penalty_overtime is None
                and self.time is not None
                and class_params.time_limit < self.time
            ):
                self.status = ResultStatus.OVER_TIME

            # modify result taking into account penalties and handicap
            if (
                class_params.penalty_controls is not None
                or class_params.penalty_overtime is not None
                or class_params.apply_handicap_rule
            ):
                # save running time if time will be modified
                self.extensions["running_time"] = self.time

                # compute penalties for overtime
                if (
                    class_params.time_limit is not None
                    and class_params.penalty_overtime is not None
                    and self.time is not None
                ):
                    # compute penalty for overtime
                    penalties_overtime = 0
                    overtime = self.time
                    while overtime > class_params.time_limit:
                        overtime -= 60
                        penalties_overtime += class_params.penalty_overtime
                    self.extensions["penalties_overtime"] = penalties_overtime
                    self.time += penalties_overtime

                # add penalties for missing controls
                if class_params.penalty_controls is not None and self.time is not None:
                    self.extensions["penalties_controls"] = penalties_controls
                    self.time += penalties_controls

                # compute total time using handicap
                if class_params.apply_handicap_rule and self.time is not None:
                    self.extensions["factor"] = handicap_factor
                    self.time = int(handicap_factor * self.time)



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import pathlib

import web
from web.utils import Storage

from ooresults.handler import model
from ooresults.utils.globals import t_globals


templates = pathlib.Path(__file__).resolve().parent.parent / "templates"
render = web.template.render(templates, globals=t_globals)


class Si1:
    def GET(self):
        id = ""
        key = ""

        try:
            for event in model.get_events():
                if event.key:
                    id = event.id
                    key = event.key
                    break
        except:
            pass

        return render.si.si1_page(id, key)


class Si2:
    def GET(self):
        id = ""
        key = ""
        data = web.input()

        try:
            for event in model.get_events():
                if str(event.id) == data.id:
                    id = event.id
                    key = event.key
        except:
            pass

        return render.si.si2_page(id, key)



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import io
from datetime import datetime
from datetime import timedelta
from typing import List
from typing import Dict

import clevercsv as csv
from unidecode import unidecode

from ooresults.repo import result_type
from ooresults.repo import start_type
from ooresults.repo.result_type import ResultStatus


def cp1252(value: str) -> str:
    try:
        _ = value.encode("windows-1252")
        return value
    except:
        return unidecode(value)


def create(entries: List[Dict], class_list: List[Dict]) -> bytes:
    output = io.StringIO()
    writer = csv.writer(output, delimiter=";", quoting=csv.QUOTE_MINIMAL)

    # write header
    writer.writerow(
        [
            "Stnr",
            "Chip",
            "Datenbank Id",
            "Nachname",
            "Vorname",
            "Jg",
            "G",
            "Block",
            "AK",
            "Start",
            "Ziel",
            "Zeit",
            "Wertung",
            "Club-Nr.",
            "Abk",
            "Ort",
            "Nat",
            "Katnr",
            "Kurz",
            "Lang",
        ]
    )

    STATUS_MAP = {
        ResultStatus.INACTIVE: "",
        ResultStatus.FINISHED: "",
        ResultStatus.OK: "0",
        ResultStatus.MISSING_PUNCH: "3",
        ResultStatus.DID_NOT_START: "1",
        ResultStatus.DID_NOT_FINISH: "2",
        ResultStatus.OVER_TIME: "5",
        ResultStatus.DISQUALIFIED: "4",
    }

    # write entries
    for i, e in enumerate(entries):
        class_no = None
        class_short_name = None
        for j, c in enumerate(class_list):
            if c.get("id") == e.get("class_id", None):
                class_no = j + 1
                class_short_name = c.get("short_name", None)
                break

        # export only items with defined name
        if e.get("last_name", None) != None:
            writer.writerow(
                [
                    str(i + 1),
                    e.get("chip", ""),
                    "",
                    cp1252(e.get("last_name", "")),
                    cp1252(e.get("first_name", "")),
                    str(e.get("year", None)) if e.get("year", None) is not None else "",
                    {"": "", "F": "W", "M": "M"}[e.get("gender", "")],
                    "",
                    "X" if e.get("not_competing", False) else "0",
                    e.result.start_time.strftime("%H:%M:%S")
                    if e.result.start_time is not None
                    else "",
                    e.result.finish_time.strftime("%H:%M:%S")
                    if e.result.finish_time is not None
                    else "",
                    "{:d}:{:02d}".format(e.result.time // 60, e.result.time % 60)
                    if e.result.time is not None
                    else "",
                    STATUS_MAP.get(e.result.status, ""),
                    str(e.get("club_id", None))
                    if e.get("club_id", None) is not None
                    else "",
                    "",
                    cp1252(e.get("club", ""))
                    if e.get("club_id", None) is not None
                    else "",
                    "",
                    str(class_no) if class_no is not None else "",
                    class_short_name
                    if class_short_name is not None
                    else e.get("class_", ""),
                    e.get("class_", ""),
                ]
            )

    content = output.getvalue()
    output.close()
    return content.encode(encoding="windows-1252")


def parse(content: bytes) -> List[Dict]:
    column_nr = {}
    result = []

    try:
        content = content.decode(encoding="utf-8")
    except:
        content = content.decode(encoding="windows-1252")

    dialect = csv.Sniffer().sniff(content, delimiters=",;\t")
    for values in csv.reader(io.StringIO(content), dialect=dialect):
        if column_nr == {}:
            for i, v in enumerate(values):
                if v in ["Chip", "SI card1"]:
                    column_nr["chip"] = i
                    break
            else:
                raise RuntimeError("Chip column not found")

            for i, v in enumerate(values):
                if v in ["First name", "Vorname"]:
                    column_nr["first_name"] = i
                    break
            else:
                raise RuntimeError("First name column not found")

            for i, v in enumerate(values):
                if v in ["Last name", "Nachname", "Surname"]:
                    column_nr["last_name"] = i
                    break
            else:
                raise RuntimeError("Last name column not found")

            for i, v in enumerate(values):
                if v in ["Time", "Zeit", "Time1"]:
                    column_nr["time"] = i
                    break
            else:
                raise RuntimeError("Time column not found")

            for i, v in enumerate(values):
                if v in ["Gender", "G", "S"]:
                    column_nr["gender"] = i
                    break
            else:
                raise RuntimeError("Gender column not found")

            for i, v in enumerate(values):
                if v in ["Year", "Jg", "YB"]:
                    column_nr["year"] = i
                    break
            else:
                raise RuntimeError("Year column not found")

            for i, v in enumerate(values):
                if v in ["AK", "nc"]:
                    column_nr["AK"] = i
                    break
            else:
                raise RuntimeError("AK column not found")

            for i, v in enumerate(values):
                if v in ["Result", "Wertung", "Classifier", "Classifier1"]:
                    column_nr["status"] = i
                    break
            else:
                raise RuntimeError("Result column not found")

            for i, v in enumerate(values):
                if v in ["Class", "Lang", "Long"]:
                    column_nr["class_"] = i
                    break
            else:
                raise RuntimeError("Class column not found")

            for i, v in enumerate(values):
                if v in ["Club", "Ort", "City"]:
                    column_nr["club"] = i
                    break
            else:
                raise RuntimeError("Club column not found")

            for i, v in enumerate(values):
                if v in ["Start", "Start1"]:
                    column_nr["start_time"] = i
                    break
            else:
                raise RuntimeError("Start column not found")

            for i, v in enumerate(values):
                if v in ["Ziel", "Finish", "Finish1"]:
                    column_nr["finish_time"] = i
                    break
            else:
                raise RuntimeError("Finish column not found")

            for i, v in enumerate(values):
                if v in ["Text1"]:
                    column_nr["text_1"] = i
                    break

            for i, v in enumerate(values):
                if v in ["Text2"]:
                    column_nr["text_2"] = i
                    break

            for i, v in enumerate(values):
                if v in ["Text3"]:
                    column_nr["text_3"] = i
                    break

            for i, v in enumerate(values):
                if v in ["Posten1", "Control1"]:
                    column_nr["split_time"] = i
                    break
        else:
            r = {
                "start": start_type.PersonRaceStart(),
                "result": result_type.PersonRaceResult(),
            }
            fields = {}

            for column in column_nr:
                item = values[column_nr[column]]
                if column == "status":
                    mapping = {
                        "0": ResultStatus.OK,
                        "1": ResultStatus.DID_NOT_START,
                        "2": ResultStatus.DID_NOT_FINISH,
                        "3": ResultStatus.MISSING_PUNCH,
                        "4": ResultStatus.DISQUALIFIED,
                        "5": ResultStatus.OVER_TIME,
                    }
                    r["result"].status = mapping.get(item, ResultStatus.INACTIVE)
                elif column == "start_time":
                    if item != "":
                        t = 0
                        for i in item.split(":"):
                            t = 60 * t + int(i)
                        d = datetime(year=1900, month=1, day=1) + timedelta(seconds=t)
                        r["start"].start_time = d
                elif column == "finish_time":
                    if item != "":
                        t = 0
                        for i in item.split(":"):
                            t = 60 * t + int(i)
                        d = datetime(year=1900, month=1, day=1) + timedelta(seconds=t)
                        r["result"].finish_time = d
                        r["result"].punched_finish_time = d
                elif column == "time":
                    try:
                        t = 0
                        for i in item.split(":"):
                            t = 60 * t + int(i)
                        r["result"].time = t
                    except:
                        pass
                elif column == "year":
                    try:
                        if int(item) <= 25:
                            r["year"] = 2000 + int(item)
                        elif int(item) <= 99:
                            r["year"] = 1900 + int(item)
                        else:
                            r["year"] = int(item)
                    except:
                        r["year"] = None
                elif column == "gender":
                    if item == "":
                        r["gender"] = ""
                    elif item in ["f", "F", "w", "W", "d", "D"]:
                        r["gender"] = "F"
                    else:
                        r["gender"] = "M"
                elif column == "AK":
                    r["not_competing"] = item == "X"
                elif column == "text_1":
                    fields[0] = item
                elif column == "text_2":
                    fields[1] = item
                elif column == "text_3":
                    fields[2] = item
                elif column == "split_time":
                    for i in range(column_nr[column], len(values) - 1, 2):
                        # posten_1, stempelzeit_1, ..., posten_n, stempelzeit_n
                        control_code = values[i]
                        control_time = values[i + 1]
                        try:
                            t = 0
                            for j in control_time.split(":"):
                                t = 60 * t + int(j)
                        except:
                            t = None

                        r["result"].split_times.append(
                            result_type.SplitTime(
                                control_code=control_code,
                                time=t,
                                status="Missing" if control_time == "-----" else None,
                            )
                        )
                else:
                    r[column] = item

            # add extra fields
            if fields:
                r["fields"] = fields

            # correct status
            if (
                r["result"].time is None
                and r["result"].finish_time is None
                and r["result"].status == ResultStatus.OK
            ):
                r["result"].status = ResultStatus.INACTIVE

            # store start time as start punch time only if finish time is defined
            # or status is missing punch, did not finished or over time
            if r["result"].finish_time is not None or r["result"].status in [
                ResultStatus.MISSING_PUNCH,
                ResultStatus.DID_NOT_FINISH,
                ResultStatus.OVER_TIME,
            ]:
                r["result"].start_time = r["start"].start_time
                r["result"].punched_start_time = r["start"].start_time

            # do not import entries with special names
            blacklist = ("Vacant", "Vakant", "Reserve")
            if r["last_name"] not in blacklist or r["first_name"] != "":
                result.append(r)

    return result



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import pathlib
from typing import List
from typing import Dict

from lxml import etree
from lxml.builder import ElementMaker


schema_file = pathlib.Path(__file__).parent.parent / "schema" / "IOF.xsd"
xml_schema = etree.XMLSchema(etree.parse(str(schema_file)))


iof_namespace = "http://www.orienteering.org/datastandard/3.0"
namespaces = {None: iof_namespace}


def create_competitor_list(competitors: List[Dict]) -> bytes:
    E = ElementMaker(namespace=iof_namespace, nsmap=namespaces)

    COMPETITORLIST = E.CompetitorList
    COMPETITOR = E.Competitor
    PERSON = E.Person
    NAME = E.Name
    FAMILY = E.Family
    GIVEN = E.Given
    BIRTHDATE = E.BirthDate
    ORGANISATION = E.Organisation
    CONTROLCARD = E.ControlCard

    root = COMPETITORLIST(iofVersion="3.0")

    for c in competitors:
        competitor = COMPETITOR()
        person = PERSON(
            NAME(
                FAMILY(c.last_name),
                GIVEN(c.first_name),
            ),
        )
        if c.gender != "":
            person.set("sex", c.gender)
        if c.year is not None:
            person.append(BIRTHDATE(str(c.year) + "-01-01"))
        competitor.append(person)
        if c.club_name is not None and c.club_name != "":
            competitor.append(ORGANISATION(NAME(c.club_name)))
        if c.chip != "":
            competitor.append(CONTROLCARD(c.chip, punchingSystem="SI"))
        root.append(competitor)

    if not xml_schema.validate(root):
        raise RuntimeError(xml_schema.error_log.last_error)
    return etree.tostring(
        root, encoding="UTF-8", xml_declaration=True, pretty_print=True
    )


def parse_competitor_list(content: bytes) -> List[Dict]:
    root = etree.XML(content)
    if not xml_schema.validate(root):
        raise RuntimeError(xml_schema.error_log.last_error)
    if not root.tag == "{" + iof_namespace + "}CompetitorList":
        raise RuntimeError(
            "Root element is " + root.tag + " but should be CompetitorList"
        )

    result = []
    competitors = root.findall("Competitor", namespaces=namespaces)
    for c in competitors:
        print(c.tag)
        for i in c:
            print(i.tag)
        r = {
            "first_name": "",
            "last_name": "",
            "club": "",
            "chip": "",
            "gender": "",
            "year": None,
        }

        r["last_name"] = c.find("Person/Name/Family", namespaces=namespaces).text
        r["first_name"] = c.find("Person/Name/Given", namespaces=namespaces).text

        e_person = c.find("Person", namespaces=namespaces)
        if e_person.get("sex") is not None:
            r["gender"] = e_person.get("sex")
        e_birthdate = c.find("Person/BirthDate", namespaces=namespaces)
        if e_birthdate is not None:
            r["year"] = int(e_birthdate.text[0:4])
        e_organization = c.find("Organisation/Name", namespaces=namespaces)
        if e_organization is not None:
            r["club"] = e_organization.text
        e_controlcard = c.find("ControlCard", namespaces=namespaces)
        if e_controlcard is not None:
            r["chip"] = e_controlcard.text

        result.append(r)

    return result
